"""
Slider widgets for the GUI.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

__all__ = ["DualHandleSlider", "SliderRow", "HistogramRangeSlider"]

import numpy as np
from qtpy.QtGui import QColor
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QWidget,
    QSlider,
    QSizePolicy,
    QHBoxLayout,
    QVBoxLayout,
    QLabel,
)

from ..stylesheets import Colors


class SliderRow(QWidget):
    """A row with label, slider, and value display."""

    valueChanged = Signal(float)

    def __init__(
        self,
        label: str,
        min_val: float = 0.0,
        max_val: float = 1.0,
        default: float = 0.5,
        decimals: int = 2,
        suffix: str = "",
        label_position: str = "left",
        steps: int = 100,
        exponent: float = 1.0,
        values: list = None,
        parent=None,
    ):
        super().__init__(parent)
        self._values = values
        if values:
            self.min_val, self.max_val = 0, len(values) - 1
            self.steps = len(values) - 1
            self.decimals = 0
        else:
            self.min_val = min_val
            self.max_val = max_val
            self.steps = steps
        self.decimals = 0 if values else decimals
        self.suffix = suffix
        self.exponent = exponent
        self._setup_ui(label, default, label_position)

    def _setup_ui(self, label: str, default: float, label_position: str = "left"):
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(12)

        self.label_widget = QLabel(f"{label}:")

        self.slider = QSlider(Qt.Orientation.Horizontal)
        self.slider.setMinimum(0)
        self.slider.setMaximum(self.steps)
        self.slider.setValue(self._value_to_slider(default))
        self.slider.valueChanged.connect(self._on_slider_changed)

        self.value_label = QLabel()
        self.value_label.setStyleSheet("QLabel { min-width: 45px; text-align: right;}")

        self.value_label.setAlignment(
            Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter
        )
        self._update_value_label(default)

        if label_position == "left":
            layout.addWidget(self.label_widget, 0, Qt.AlignmentFlag.AlignVCenter)
            layout.addWidget(self.slider, 1, Qt.AlignmentFlag.AlignVCenter)
            layout.addWidget(self.value_label, 0, Qt.AlignmentFlag.AlignVCenter)
        else:
            layout.addWidget(self.slider, 1, Qt.AlignmentFlag.AlignVCenter)
            layout.addWidget(self.label_widget, 0, Qt.AlignmentFlag.AlignVCenter)
            layout.addWidget(self.value_label, 0, Qt.AlignmentFlag.AlignVCenter)

    def _value_to_slider(self, value: float) -> int:
        """Convert actual value to slider position (0-steps)."""
        if self._values:
            try:
                return self._values.index(value)
            except ValueError:
                return 0
        ratio = (value - self.min_val) / (self.max_val - self.min_val)
        if self.exponent != 1.0:
            ratio = 1.0 - (1.0 - ratio) ** (1.0 / self.exponent)
        return int(ratio * self.steps)

    def _slider_to_value(self, pos: int) -> float:
        """Convert slider position to actual value."""
        if self._values:
            return self._values[min(pos, len(self._values) - 1)]
        ratio = pos / self.steps
        if self.exponent != 1.0:
            ratio = 1.0 - (1.0 - ratio) ** self.exponent
        return self.min_val + ratio * (self.max_val - self.min_val)

    def _update_value_label(self, value: float):
        """Update the value label display."""
        if self.decimals == 0:
            text = f"{int(value)}{self.suffix}"
        else:
            text = f"{value:.{self.decimals}f}{self.suffix}"
        self.value_label.setText(text)

    def _on_slider_changed(self, pos: int):
        """Handle slider value change."""
        value = self._slider_to_value(pos)
        self._update_value_label(value)
        self.valueChanged.emit(value)

    def value(self) -> float:
        """Get the current value."""
        return self._slider_to_value(self.slider.value())

    def setValue(self, value: float):
        """Set the current value."""
        self.slider.setValue(self._value_to_slider(value))
        self._update_value_label(value)

    def setRange(self, min_val: float, max_val: float):
        """Set the value range."""
        self.min_val = min_val
        self.max_val = max_val
        if self.decimals == 0:
            self.steps = max(1, int(max_val - min_val))
            self.slider.setMaximum(self.steps)
        self._update_value_label(self.value())

    def setEnabled(self, enabled: bool):
        """Enable or disable the widget."""
        super().setEnabled(enabled)
        self.label_widget.setEnabled(enabled)
        self.slider.setEnabled(enabled)
        self.value_label.setEnabled(enabled)


class DualHandleSlider(QWidget):
    """A slider with two handles for selecting a range, with visual feedback."""

    rangeChanged = Signal(float, float)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.min_val = 0.0
        self.max_val = 100.0
        self.lower_pos = 0.0
        self.upper_pos = 100.0

        self.setMinimumHeight(40)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # Track which handle is being dragged
        self.dragging_handle = None
        self.handle_size = 16

        # Colors matching QSlider_style from stylesheets.py
        self.groove_color = QColor("#e2e8f0")
        self.active_color = QColor(Colors.BORDER_HOVER)
        self.handle_color = QColor("#ffffff")
        self.border_color = QColor(Colors.BORDER_DARK)
        # Disabled colors
        self.groove_disabled = QColor(Colors.BG_TERTIARY)
        self.active_disabled = QColor(Colors.BORDER_DARK)
        self.handle_disabled = QColor(Colors.BG_SECONDARY)
        self.border_disabled = QColor("#e2e8f0")

    def setRange(self, minimum, maximum):
        """Set the range of values the slider represents."""
        self.min_val = minimum
        self.max_val = maximum
        self.update()

    def setValues(self, lower, upper):
        """Set both handle positions."""
        self.lower_pos = np.clip(lower, self.min_val, self.max_val)
        self.upper_pos = np.clip(upper, self.min_val, self.max_val)
        if self.lower_pos > self.upper_pos:
            self.lower_pos, self.upper_pos = self.upper_pos, self.lower_pos
        self.update()

    def _value_to_pixel(self, value):
        """Convert a value to pixel position."""
        if self.max_val == self.min_val:
            return self.handle_size
        margin = self.handle_size
        width = self.width() - margin * 2
        normalized = (value - self.min_val) / (self.max_val - self.min_val)
        return margin + normalized * width

    def _pixel_to_value(self, pixel):
        """Convert pixel position to value."""
        margin = self.handle_size
        width = self.width() - margin * 2
        if width <= 0:
            return self.min_val
        normalized = (pixel - margin) / width
        normalized = np.clip(normalized, 0, 1)
        return self.min_val + normalized * (self.max_val - self.min_val)

    def paintEvent(self, event):
        """Draw the slider with range visualization."""
        from qtpy.QtGui import QPainter, QPen, QBrush

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        # Use disabled colors when widget is disabled
        if not self.isEnabled():
            groove = self.groove_disabled
            active = self.active_disabled
            handle = self.handle_disabled
            border = self.border_disabled
        else:
            groove = self.groove_color
            active = self.active_color
            handle = self.handle_color
            border = self.border_color

        # Draw track/groove
        track_y = self.height() // 2
        margin = self.handle_size
        track_width = self.width() - margin * 2

        painter.setPen(Qt.NoPen)
        painter.setBrush(QBrush(groove))
        painter.drawRoundedRect(margin, track_y - 2, track_width, 4, 2, 2)

        # Draw active range
        lower_x = self._value_to_pixel(self.lower_pos)
        upper_x = self._value_to_pixel(self.upper_pos)
        range_width = upper_x - lower_x

        painter.setBrush(QBrush(active))
        painter.drawRoundedRect(int(lower_x), track_y - 2, int(range_width), 4, 2, 2)

        # Draw handles (circular, matching QSlider style)
        for pos in [self.lower_pos, self.upper_pos]:
            x = self._value_to_pixel(pos)
            handle_x = int(x - self.handle_size // 2)
            handle_y = track_y - self.handle_size // 2

            painter.setBrush(QBrush(handle))
            painter.setPen(QPen(border, 1))
            painter.drawEllipse(handle_x, handle_y, self.handle_size, self.handle_size)

    def mousePressEvent(self, event):
        """Start dragging a handle."""
        if not self.isEnabled() or event.button() != Qt.LeftButton:
            return

        x = event.pos().x()
        lower_x = self._value_to_pixel(self.lower_pos)
        upper_x = self._value_to_pixel(self.upper_pos)

        # Check which handle is closer
        dist_to_lower = abs(x - lower_x)
        dist_to_upper = abs(x - upper_x)

        if dist_to_lower < self.handle_size:
            self.dragging_handle = "lower"
        elif dist_to_upper < self.handle_size:
            self.dragging_handle = "upper"
        else:
            # Click on track - move nearest handle
            if dist_to_lower < dist_to_upper:
                self.dragging_handle = "lower"
                self.lower_pos = self._pixel_to_value(x)
            else:
                self.dragging_handle = "upper"
                self.upper_pos = self._pixel_to_value(x)
            self.update()
            self.rangeChanged.emit(self.lower_pos, self.upper_pos)

    def mouseMoveEvent(self, event):
        """Drag the active handle."""
        if not self.isEnabled() or self.dragging_handle is None:
            return

        value = self._pixel_to_value(event.pos().x())

        if self.dragging_handle == "lower":
            self.lower_pos = min(value, self.upper_pos)
        else:
            self.upper_pos = max(value, self.lower_pos)

        self.update()
        self.rangeChanged.emit(self.lower_pos, self.upper_pos)

    def mouseReleaseEvent(self, event):
        """Stop dragging."""
        was_dragging = self.dragging_handle is not None
        if event.button() == Qt.LeftButton:
            self.dragging_handle = None
            if was_dragging:
                self.rangeReleased.emit(self.lower_pos, self.upper_pos)

    rangeReleased = Signal(float, float)


class MiniHistogram(QWidget):
    """A compact histogram display widget."""

    def __init__(self, parent=None, n_bins: int = 50, margin: int = 16):
        super().__init__(parent)
        self.n_bins = n_bins
        self.margin = margin  # Match DualHandleSlider handle_size for alignment
        self._hist = None
        self._bin_edges = None
        self._lower = 0.0
        self._upper = 1.0
        self._min_val = 0.0
        self._max_val = 1.0

        self.hist_color = QColor(Colors.BORDER_HOVER)
        self.hist_color.setAlpha(100)
        self.selected_color = QColor(Colors.PRIMARY)
        self.selected_color.setAlpha(160)
        self.placeholder_color = QColor(Colors.BORDER_DARK)
        self.placeholder_color.setAlpha(30)

    def setData(self, values):
        """Compute histogram from values."""
        if values is None or len(values) == 0:
            self._hist = None
            self.update()
            return

        values = np.asarray(values).flatten()
        values = values[np.isfinite(values)]
        if len(values) == 0:
            self._hist = None
            return

        self._min_val = float(np.min(values))
        self._max_val = float(np.max(values))
        if self._min_val == self._max_val:
            self._max_val = self._min_val + 1.0

        self._hist, self._bin_edges = np.histogram(values, bins=self.n_bins)
        self._lower = self._min_val
        self._upper = self._max_val
        self.update()

    def setSelection(self, lower, upper):
        """Update the highlighted range."""
        self._lower = lower
        self._upper = upper
        self.update()

    def paintEvent(self, event):
        """Draw the histogram bars or placeholder."""
        from qtpy.QtGui import QPainter, QBrush, QPen

        painter = QPainter(self)
        width = self.width()
        height = self.height()
        draw_width = width - 2 * self.margin

        if self._hist is None or len(self._hist) == 0:
            # Draw placeholder
            painter.setPen(Qt.NoPen)
            painter.setBrush(QBrush(self.placeholder_color))
            painter.drawRect(self.margin, 0, draw_width, height)

            painter.setPen(QPen(QColor(Colors.TEXT_SECONDARY)))
            painter.drawText(
                self.margin, 0, draw_width, height, Qt.AlignCenter, "No data"
            )
            return

        max_count = np.max(self._hist)
        if max_count == 0:
            return

        bar_width = draw_width / len(self._hist)
        for i, count in enumerate(self._hist):
            bar_height = int((count / max_count) * (height - 2))
            x = self.margin + i * bar_width
            y = height - bar_height

            bin_center = (self._bin_edges[i] + self._bin_edges[i + 1]) / 2
            if self._lower <= bin_center <= self._upper:
                painter.setBrush(QBrush(self.selected_color))
            else:
                painter.setBrush(QBrush(self.hist_color))

            painter.setPen(Qt.NoPen)
            painter.drawRect(int(x), int(y), int(bar_width - 1), bar_height)


class HistogramRangeSlider(QWidget):
    """A histogram with an integrated range slider for filtering."""

    rangeChanged = Signal(float, float)
    rangeReleased = Signal(float, float)

    def __init__(self, parent=None, n_bins: int = 50):
        super().__init__(parent)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)

        self._histogram = MiniHistogram(n_bins=n_bins)
        self._slider = DualHandleSlider()

        layout.addWidget(self._histogram)
        layout.addWidget(self._slider)

        self._slider.rangeChanged.connect(self._on_range_changed)
        self._slider.rangeReleased.connect(self.rangeReleased)

    def _on_range_changed(self, lower, upper):
        self._histogram.setSelection(lower, upper)
        self.rangeChanged.emit(lower, upper)

    def setData(self, values):
        """Set data and update histogram and slider range."""
        self._histogram.setData(values)
        if values is not None and len(values) > 0:
            min_val = self._histogram._min_val
            max_val = self._histogram._max_val
            self._slider.setRange(min_val, max_val)
            self._slider.setValues(min_val, max_val)

    def getRange(self):
        """Get the current selected range."""
        return self._slider.lower_pos, self._slider.upper_pos



"""
Main pipeline builder dialog.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import json
import uuid

from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QWidget,
    QGroupBox,
    QScrollArea,
    QSpinBox,
    QFileDialog,
    QMessageBox,
    QCheckBox,
)
from qtpy.QtCore import Qt, QSize
from qtpy.QtGui import QFont

from .executor import generate_runs
from ._utils import natural_sort_key, strip_filepath
from .operations import OPERATION_CATEGORIES, PIPELINE_PRESETS
from .widgets import OperationCardWidget, PipelineTreeWidget

from ..__version__ import __version__
from ..settings import Settings
from ..widgets.settings import format_tooltip
from ..widgets import SearchWidget, ContainerListWidget
from ..widgets.container_list import StyledTreeWidgetItem
from ..stylesheets import Colors, QPushButton_style, QScrollArea_style


__all__ = ["BatchNavigatorDialog", "PipelineBuilderDialog"]


class PipelineBuilderDialog(QDialog):
    """Pipeline builder with tree structure and operation library."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Pipeline Builder")
        self.setMinimumSize(1100, 750)
        self.resize(1300, 850)
        self.setup_ui()
        self.setStyleSheet(QScrollArea_style + QPushButton_style)

    def setup_ui(self):
        import qtawesome as qta
        from ..icons import dialog_accept_icon, dialog_reject_icon

        self.pipeline_tree = PipelineTreeWidget()
        self.pipeline_tree.pipeline_changed.connect(self._update_library_state)

        main_layout = QVBoxLayout(self)
        main_layout.setSpacing(12)
        main_layout.setContentsMargins(12, 12, 12, 12)

        content_splitter = QWidget()
        content_layout = QHBoxLayout(content_splitter)
        content_layout.setSpacing(12)
        content_layout.setContentsMargins(0, 0, 0, 0)

        library_group = QGroupBox("Operation Library")
        library_group.setFixedWidth(280)
        library_layout = QVBoxLayout()
        library_layout.setContentsMargins(8, 8, 8, 8)
        library_layout.setSpacing(0)

        library_scroll = QScrollArea()
        library_scroll.setWidgetResizable(True)
        library_scroll.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)

        self.library_widget = self._create_library()
        library_scroll.setWidget(self.library_widget)
        library_layout.addWidget(library_scroll)

        library_group.setLayout(library_layout)
        content_layout.addWidget(library_group)

        workflow_group = QGroupBox("Pipeline Workflow")
        workflow_layout = QVBoxLayout()
        workflow_layout.setContentsMargins(8, 8, 8, 8)
        workflow_layout.setSpacing(8)

        info_label = QLabel("Operations to execute in sequence")
        info_label.setStyleSheet(f"color: {Colors.TEXT_MUTED}; font-size: 11px;")
        info_label.setWordWrap(True)
        workflow_layout.addWidget(info_label)

        workflow_layout.addWidget(self.pipeline_tree, 1)

        workflow_group.setLayout(workflow_layout)
        content_layout.addWidget(workflow_group, 1)

        main_layout.addWidget(content_splitter, 1)

        settings_row = QHBoxLayout()
        settings_row.setSpacing(12)

        presets_group = QGroupBox("Presets")
        presets_container = QVBoxLayout()
        presets_container.setSpacing(8)

        presets_label = QLabel("Common Workflow Configurations:")
        presets_label.setStyleSheet(f"font-size: 11px; color: {Colors.TEXT_MUTED};")
        presets_container.addWidget(presets_label)

        presets_layout = QHBoxLayout()

        preset_buttons = [
            ("Clear", "ph.x-circle", Colors.CATEGORY["clear"]),
            ("Import", "ph.file-arrow-up", Colors.CATEGORY["import"]),
            ("Cleanup", "ph.wrench", Colors.CATEGORY["cleanup"]),
            ("Meshing", "ph.triangle", Colors.CATEGORY["meshing"]),
            ("Particle Picking", "ph.crosshair", Colors.CATEGORY["particle_picking"]),
        ]

        self.preset_buttons = {}
        for idx, (name, icon, color) in enumerate(preset_buttons):
            btn = QPushButton(name)
            btn.setIcon(qta.icon(icon, color=color))
            btn.setIconSize(QSize(24, 24))
            btn.setFixedHeight(32)

            preset_key = name.replace("\n", " ")
            btn.clicked.connect(lambda checked, n=preset_key: self._load_preset(n))
            presets_layout.addWidget(btn, 0)
            self.preset_buttons[preset_key] = btn

        presets_container.addLayout(presets_layout)
        presets_container.addStretch()
        presets_group.setLayout(presets_container)

        workers_group = QGroupBox("Settings")
        workers_layout = QHBoxLayout()
        workers_layout.setSpacing(8)

        workers_vbox = QVBoxLayout()
        workers_vbox.setSpacing(4)
        workers_label = QLabel("Parallel Workers:")
        workers_label.setToolTip(
            format_tooltip(
                label="Parallel Workers",
                description="Number of parallel processes in the pipeline.",
                notes="Depending on the workflow, i.e., number of steps and how many "
                "outputs are saved, calculate approximately 8GB of RAM per worker.",
            )
        )
        workers_label.setStyleSheet(f"font-size: 11px; color: {Colors.TEXT_MUTED};")
        self.workers_spin = QSpinBox()
        self.workers_spin.setMinimum(1)
        self.workers_spin.setMaximum(Settings.rendering.parallel_worker)
        self.workers_spin.setValue(
            int(getattr(Settings.rendering, "parallel_worker", 4))
        )
        self.workers_spin.setFixedWidth(80)
        self.workers_spin.setFixedHeight(32)
        workers_vbox.addWidget(workers_label)
        workers_vbox.addWidget(self.workers_spin)

        skip_vbox = QVBoxLayout()
        skip_vbox.setSpacing(4)
        skip_complete_label = QLabel("Skip Complete:")
        skip_complete_label.setToolTip(
            format_tooltip(
                label="Skip Complete",
                description="Skip runs where output files already exist.",
            )
        )
        skip_complete_label.setStyleSheet(
            f"font-size: 11px; color: {Colors.TEXT_MUTED};"
        )
        self.skip_complete = QCheckBox()
        self.skip_complete.setFixedHeight(32)
        skip_vbox.addWidget(skip_complete_label)
        skip_vbox.addWidget(self.skip_complete)

        workers_layout.addLayout(workers_vbox)
        workers_layout.addLayout(skip_vbox)
        workers_layout.addStretch()

        workers_group.setLayout(workers_layout)

        settings_row.addWidget(presets_group, 1)
        settings_row.addWidget(workers_group)
        main_layout.addLayout(settings_row)

        footer_layout = QHBoxLayout()
        footer_layout.setSpacing(6)

        load_btn = QPushButton("Load Pipeline")
        load_btn.clicked.connect(self._load_config)
        load_btn.setIcon(qta.icon("ph.upload", color=Colors.PRIMARY))
        footer_layout.addWidget(load_btn)

        export_btn = QPushButton("Export Pipeline")
        export_btn.clicked.connect(self._export_config)
        export_btn.setIcon(qta.icon("ph.download", color=Colors.PRIMARY))
        footer_layout.addWidget(export_btn)

        validate_btn = QPushButton("Validate Pipeline")
        validate_btn.clicked.connect(self._validate_pipeline)
        validate_btn.setIcon(qta.icon("ph.check", color=Colors.PRIMARY))
        footer_layout.addWidget(validate_btn)

        footer_layout.addStretch()

        cancel_btn = QPushButton("Cancel")
        cancel_btn.setIcon(dialog_reject_icon)
        cancel_btn.clicked.connect(self.reject)
        footer_layout.addWidget(cancel_btn)

        run_btn = QPushButton("Run Pipeline")
        run_btn.clicked.connect(self.accept)
        run_btn.setIcon(dialog_accept_icon)
        run_btn.setDefault(True)
        footer_layout.addWidget(run_btn)

        main_layout.addLayout(footer_layout)

    def _create_library(self):
        """Create operation library widget with categories."""

        widget = QWidget()
        layout = QVBoxLayout(widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(8)

        info_label = QLabel("Select operations to add to your pipeline")
        info_label.setStyleSheet(f"color: {Colors.TEXT_MUTED}; font-size: 11px;")
        info_label.setWordWrap(True)
        layout.addWidget(info_label)

        self._library_buttons = {}
        for category_id, category in OPERATION_CATEGORIES.items():
            cat_header = QLabel(category["title"])
            cat_header_font = QFont()
            cat_header_font.setBold(True)
            cat_header_font.setPointSize(11)
            cat_header.setFont(cat_header_font)
            cat_header.setStyleSheet(f"color: {category['color']}; margin-top: 8px;")
            layout.addWidget(cat_header)

            for op_name, op_info in category["operations"].items():
                op_info["category"] = category_id
                op_btn = self._create_library_button(
                    op_name, op_info, category["color"]
                )
                op_btn.clicked.connect(
                    lambda checked, n=op_name, i=op_info, c=category[
                        "color"
                    ]: self._add_card(n, i, c)
                )
                layout.addWidget(op_btn)
                self._library_buttons[(category_id, op_name)] = op_btn

        layout.addStretch()
        self._update_library_state()
        return widget

    def _create_library_button(self, name, info, color):
        """Create compact button for library operation."""
        import qtawesome as qta

        btn = QPushButton()
        btn.setFixedHeight(50)
        btn.setCursor(Qt.CursorShape.PointingHandCursor)

        base_style = f"""
            QPushButton {{
                border: 1px solid {Colors.BORDER_DARK};
                border-left: 3px solid {color};
                border-radius: 4px;
                text-align: left;
                padding: 6px;
            }}
            QPushButton:hover {{
                background: {Colors.BG_SECONDARY};
                border-left-color: {color};
            }}
        """
        btn.setStyleSheet(base_style)
        btn.setProperty("base_style", base_style)

        btn_layout = QHBoxLayout()
        btn_layout.setSpacing(6)
        btn_layout.setContentsMargins(4, 4, 4, 4)

        icon_label = QLabel()
        icon_label.setPixmap(qta.icon(info["icon"], color=color).pixmap(20, 20))
        btn_layout.addWidget(icon_label)

        text_layout = QVBoxLayout()
        text_layout.setSpacing(2)

        name_label = QLabel(name)
        name_label.setStyleSheet(f"color: {color}; font-weight: 600; font-size: 11px;")
        text_layout.addWidget(name_label)

        desc_label = QLabel(info["description"])
        desc_label.setStyleSheet(f"color: {Colors.TEXT_MUTED}; font-size: 11px;")
        desc_label.setWordWrap(True)

        text_layout.addWidget(desc_label)
        btn_layout.addLayout(text_layout, 1)

        add_icon_label = QLabel()
        add_icon_label.setPixmap(
            qta.icon("ph.plus", color=Colors.ICON_MUTED).pixmap(14, 14)
        )
        btn_layout.addWidget(add_icon_label)

        btn.setLayout(btn_layout)
        return btn

    def _get_current_output_type(self):
        """Get the output type of the last operation in the pipeline."""
        if self.pipeline_tree.topLevelItemCount() == 0:
            return None

        # Find the last operation card that has an output type
        for i in range(self.pipeline_tree.topLevelItemCount() - 1, -1, -1):
            widget = self.pipeline_tree.itemWidget(
                self.pipeline_tree.topLevelItem(i), 0
            )
            if isinstance(widget, OperationCardWidget):
                output_type = widget.operation_info.get("output_type")
                if output_type is not None:
                    return output_type
        return None

    def _is_operation_valid(self, operation_info):
        """Check if an operation can be added to the current pipeline."""
        current_output = self._get_current_output_type()
        operation_input = operation_info.get("input_type")

        # If pipeline is empty, only allow operations with no input requirement
        if current_output is None:
            return operation_input is None

        # If operation accepts any input
        if operation_input == "any":
            return True

        # If operation has no input requirement (like some export operations)
        if operation_input is None:
            return True

        # If current output is 'any', it should be compatible with 'point' operations
        # since Import Files outputs 'any' and we expect to process point clouds next
        if current_output == "any":
            return True
        return operation_input == current_output

    def _update_library_state(self):
        """Update library buttons to show which operations are valid."""

        if not hasattr(self, "_library_buttons"):
            return None

        for (category_id, op_name), btn in self._library_buttons.items():
            op_info = OPERATION_CATEGORIES[category_id]["operations"][op_name]
            is_valid = self._is_operation_valid(op_info)

            btn.setEnabled(is_valid)

            if not is_valid:
                btn.setStyleSheet(
                    btn.property("base_style")
                    + f"""
                    QPushButton:disabled {{
                        background: {Colors.BG_TERTIARY};
                    }}
                """
                )

    def _add_card(self, name, info, color, node_id=None, settings=None):
        settings = settings if isinstance(settings, dict) else {}

        previous_node = None
        count = self.pipeline_tree.topLevelItemCount()
        if count > 0:
            last_item = self.pipeline_tree.topLevelItem(count - 1)
            last_widget = self.pipeline_tree.itemWidget(last_item, 0)
            if isinstance(last_widget, OperationCardWidget):
                previous_node = last_widget.node_id

        card = OperationCardWidget(name, info, color, node_id=node_id)
        if "inputs" not in settings:
            settings["inputs"] = [previous_node] if previous_node is not None else []

        card.set_settings(settings)
        self.pipeline_tree.add_operation_card(card)

    def _load_preset(self, preset_name):
        """Load a predefined preset pipeline."""
        self.pipeline_tree.clear()

        self._update_library_state()
        if preset_name not in PIPELINE_PRESETS:
            return None

        for op in PIPELINE_PRESETS[preset_name]:
            op = op.copy()
            op_name, op_category = op["name"], op["category"]
            op_info = OPERATION_CATEGORIES[op_category]["operations"][op_name]
            self._add_card(
                op_name,
                op_info,
                OPERATION_CATEGORIES[op_category]["color"],
                settings=op,
            )

    def _load_config(self):
        """Load pipeline configuration from file (supports graph format)."""
        filename, _ = QFileDialog.getOpenFileName(
            self, "Load Pipeline Configuration", "", "JSON Files (*.json)"
        )

        if not filename:
            return None

        try:
            with open(filename, "r") as f:
                config = json.load(f)
        except (json.JSONDecodeError, IOError) as e:
            QMessageBox.warning(
                self, "Load Error", f"Failed to load configuration: {str(e)}"
            )
            return None

        nodes = config["nodes"]
        self.pipeline_tree.clear()

        self._update_library_state()
        total_ops, valid_ops = len(nodes), 0
        for node in nodes:
            category = node.get("category")
            if category not in OPERATION_CATEGORIES:
                continue

            op_name = node.get("name")
            if op_name not in OPERATION_CATEGORIES[category]["operations"]:
                continue

            op_info = OPERATION_CATEGORIES[category]["operations"][op_name]
            node_id = node.get("id", str(uuid.uuid4()))

            self._add_card(
                op_name,
                op_info,
                OPERATION_CATEGORIES[category]["color"],
                node_id=node_id,
                settings=node,
            )
            valid_ops += 1

        if "workers" in config:
            self.workers_spin.setValue(config["workers"])

        if "skip_complete" in config:
            self.skip_complete.setChecked(bool(config["skip_complete"]))

        if total_ops != valid_ops:
            QMessageBox.information(
                self,
                "Import Failed",
                f"Imported {valid_ops} of {total_ops} operations.",
            )

    def _export_config(self):
        """Export current pipeline configuration to file (graph format)."""
        filename, _ = QFileDialog.getSaveFileName(
            self, "Export Pipeline Configuration", "", "JSON Files (*.json)"
        )

        if not filename:
            return None

        try:
            config = self.get_pipeline_config()
            with open(filename, "w") as f:
                json.dump(config, f, indent=2)

            QMessageBox.information(
                self, "Export Success", f"Configuration exported to {filename}"
            )
        except IOError as e:
            QMessageBox.warning(
                self, "Export Error", f"Failed to export configuration: {str(e)}"
            )

    def _validate_pipeline(self):
        """Validate the pipeline configuration."""
        try:
            config = self.get_pipeline_config()
            runs = generate_runs(config)

            QMessageBox.information(
                self,
                "Validation Success",
                f"Pipeline is valid!\n\n"
                f"- {len(config['nodes'])} operations\n"
                f"- {len(runs)} runs will be generated\n"
                f"- No cycles detected",
            )
        except Exception as e:
            QMessageBox.warning(
                self, "Validation Failed", f"Pipeline validation failed:\n\n{str(e)}"
            )

    def get_settings(self):
        return {
            "runs": getattr(self, "pipeline_runs", []),
            "workers": self.workers_spin.value(),
            "skip_complete": self.skip_complete.isChecked(),
        }

    def accept(self):
        """Execute the pipeline on all input files."""
        try:
            config = self.get_pipeline_config()
            self.pipeline_runs = generate_runs(config)
        except Exception as e:
            QMessageBox.warning(
                self, "Pipeline Error", f"Failed to generate runs:\n\n{str(e)}"
            )
            return None

        super().accept()

    def get_pipeline_config(self):
        """Get complete pipeline configuration in graph format."""
        return {
            "version": __version__,
            "format": "directed_graph",
            "nodes": self.pipeline_tree.get_pipeline_config(),
            "workers": self.workers_spin.value(),
            "skip_complete": self.skip_complete.isChecked(),
            "metadata": {
                "description": "Mosaic batch pipeline configuration",
                "created_with": "Mosaic Pipeline Builder",
            },
        }


class BatchNavigatorDialog(QWidget):
    """Widget for navigating through batch-created sessions."""

    def __init__(self, main_window, parent=None):
        super().__init__(parent)
        self.main_window = main_window
        self.current_index = -1

        self._session_modified = False
        self.setWindowTitle("Batch Navigator")

        self.session_files = []

        self.setup_ui()
        self.setStyleSheet(QPushButton_style + QScrollArea_style)

        if hasattr(main_window, "cdata"):
            main_window.cdata.data.data_changed.connect(self._mark_modified)
            main_window.cdata.models.data_changed.connect(self._mark_modified)

    def _mark_modified(self):
        """Mark current session as modified."""
        self._session_modified = True

    def setup_ui(self):
        import qtawesome as qta

        self.setMinimumWidth(280)

        layout = QVBoxLayout(self)
        layout.setSpacing(10)
        layout.setContentsMargins(10, 10, 10, 10)

        count = len(self.session_files)
        self.sessions_group = QGroupBox(f"{count} Session{'s' if count != 1 else ''}")
        sessions_layout = QVBoxLayout(self.sessions_group)
        sessions_layout.setSpacing(8)

        self.search_widget = SearchWidget(placeholder="Search sessions...")
        self.search_widget.searchTextChanged.connect(self._filter_sessions)

        self.session_list = ContainerListWidget(border=False)
        self.session_list.tree_widget.setSelectionMode(
            self.session_list.tree_widget.SelectionMode.SingleSelection
        )
        self.session_list.tree_widget.itemClicked.connect(self._on_item_clicked)

        self._populate_session_list()

        self.auto_save_checkbox = QCheckBox("Auto-save when switching")
        self.auto_save_checkbox.setChecked(True)
        self.auto_save_checkbox.setToolTip(
            "Automatically save the current session when switching to another"
        )

        # Top bar with search and action buttons
        top_bar = QHBoxLayout()
        top_bar.setSpacing(8)
        top_bar.setContentsMargins(0, 0, 0, 0)
        top_bar.addWidget(self.search_widget, 1)

        self.add_btn = QPushButton()
        self.add_btn.setIcon(qta.icon("ph.file-arrow-up", color=Colors.ICON))
        self.add_btn.setToolTip("Add session files to the list")
        self.add_btn.clicked.connect(self._add_sessions)
        self.add_btn.setFixedSize(32, 32)

        self.clear_btn = QPushButton()
        self.clear_btn.setIcon(qta.icon("ph.x", color=Colors.ICON))
        self.clear_btn.setToolTip("Remove all sessions from the list")
        self.clear_btn.clicked.connect(self._clear_sessions)
        self.clear_btn.setFixedSize(32, 32)

        top_bar.addWidget(self.add_btn)
        top_bar.addWidget(self.clear_btn)

        sessions_layout.addLayout(top_bar)
        sessions_layout.addWidget(self.session_list, 1)
        sessions_layout.addWidget(self.auto_save_checkbox)
        layout.addWidget(self.sessions_group, 1)

        button_layout = QHBoxLayout()
        button_layout.setSpacing(8)

        self.discard_btn = QPushButton("Reload")
        self.discard_btn.setIcon(
            qta.icon("ph.arrow-counter-clockwise", color=Colors.ICON)
        )
        self.discard_btn.clicked.connect(self._discard_changes)
        self.discard_btn.setToolTip(
            "Reload current session, discarding unsaved changes"
        )

        self.save_btn = QPushButton("Save")
        self.save_btn.setToolTip("Save current session to disk")
        self.save_btn.setIcon(qta.icon("ph.floppy-disk", color=Colors.ICON))
        self.save_btn.clicked.connect(self._save_current)

        button_layout.addWidget(self.discard_btn, 1)
        button_layout.addWidget(self.save_btn, 1)
        layout.addLayout(button_layout)

    def _populate_session_list(self):
        """Populate the session list widget."""
        self.session_list.tree_widget.clear()

        for i, filepath in enumerate(self.session_files):
            item = StyledTreeWidgetItem(
                strip_filepath(filepath),
                visible=(i == self.current_index),
                metadata={"index": i, "filepath": filepath},
            )
            self.session_list.tree_widget.addTopLevelItem(item)

        count = len(self.session_files)
        self.sessions_group.setTitle(f"{count} Session{'s' if count != 1 else ''}")

    def _reset_selection(self):
        self.current_index = -1
        self.session_list.tree_widget.clearSelection()
        self._update_session_list()

    def _remove_session_at_index(self, index):
        """Remove a session from the list."""
        if not (0 <= index < len(self.session_files)):
            return

        self.session_files.pop(index)

        # If we removed the current session, reset
        if index == self.current_index:
            self._reset_selection()
        # If we removed a session before the current, adjust index
        elif index < self.current_index:
            self.current_index -= 1
            self._populate_session_list()
        else:
            self._populate_session_list()

    def _update_session_list(self):
        """Update visibility state of items in the list."""
        for i in range(self.session_list.tree_widget.topLevelItemCount()):
            item = self.session_list.tree_widget.topLevelItem(i)
            item.set_visible(i == self.current_index)

    def _filter_sessions(self, search_text):
        """Filter session list based on search text."""
        search_text = search_text.lower()

        for i in range(self.session_list.tree_widget.topLevelItemCount()):
            item = self.session_list.tree_widget.topLevelItem(i)
            filename = strip_filepath(item.metadata.get("filepath", "")).lower()

            matches = search_text in filename if search_text else True
            item.setHidden(not matches)

    def _on_item_clicked(self, item):
        """Handle single click on session item"""
        index = item.metadata.get("index", -1)
        if index >= 0 and index != self.current_index:
            self._switch_to_session(index)

    def _load_session_at_index(self, index):
        """Load a session file at the given index."""
        if not 0 <= index < len(self.session_files):
            return

        filepath = self.session_files[index]
        self.main_window._load_session(filepath)
        self.current_index = index

        self._update_session_list()

    def _switch_to_session(self, new_index):
        """Switch to a different session, auto-saving current one if enabled."""
        if new_index == self.current_index:
            return

        if self.auto_save_checkbox.isChecked():
            self._save_current()
        self._load_session_at_index(new_index)
        self._session_modified = False

    def _save_current(self):
        """Save the currently loaded session."""
        if self.current_index < 0:
            return None

        if self._session_modified:
            filepath = self.session_files[self.current_index]
            self.main_window.cdata.to_file(filepath)

    def _discard_changes(self):
        """Discard changes by reloading the current session."""
        if self.current_index < 0:
            return

        reply = QMessageBox.question(
            self,
            "Discard Changes",
            "Reload the current session and discard all unsaved changes?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )

        if reply == QMessageBox.StandardButton.Yes:
            self._load_session_at_index(self.current_index)

    def _clear_sessions(self):
        """Remove all sessions from the list."""
        self._save_current()
        self.session_files.clear()
        self.current_index = -1
        self._populate_session_list()

    def _add_sessions(self):
        """Add session files via file dialog, deduplicating paths."""
        filepaths, _ = QFileDialog.getOpenFileNames(
            self, "Add Session Files", "", "Pickle Files (*.pickle)"
        )
        if not filepaths:
            return

        existing = set(self.session_files)
        for fp in filepaths:
            if fp not in existing:
                self.session_files.append(fp)
                existing.add(fp)

        self.session_files.sort(key=natural_sort_key)
        self._populate_session_list()

    def close(self):
        """Handle widget close, saving current session if auto-save is enabled."""
        self._save_current()
        super().close()



"""
Segmented button control — mutually exclusive toggle group.

Copyright (c) 2026 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget, QHBoxLayout, QPushButton

from ..stylesheets import Colors


class SegmentedControl(QWidget):
    """Inline mutually exclusive button group.

    Parameters
    ----------
    options : list of str
        Button labels.
    default : int, optional
        Initially selected index, by default 0.
    parent : QWidget, optional
        Parent widget.
    """

    selectionChanged = Signal(str)

    def __init__(self, options, default=0, parent=None):
        super().__init__(parent)
        self._buttons = []
        self._selected = default

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        n = len(options)
        for i, text in enumerate(options):
            btn = QPushButton(text)
            btn.setCheckable(True)
            btn.setChecked(i == default)
            rl = "4px" if i == 0 else "0px"
            rr = "4px" if i == n - 1 else "0px"
            ml = "0px" if i == 0 else "-1px"
            btn.setStyleSheet(
                f"""
                QPushButton {{
                    border: 1px solid {Colors.BORDER_DARK};
                    border-radius: 0px;
                    border-top-left-radius: {rl};
                    border-bottom-left-radius: {rl};
                    border-top-right-radius: {rr};
                    border-bottom-right-radius: {rr};
                    padding: 6px 12px;
                    font-size: 12px;
                    color: {Colors.TEXT_SECONDARY};
                    background: transparent;
                    margin-left: {ml};
                }}
                QPushButton:checked {{
                    background: {Colors.BG_TERTIARY};
                    color: {Colors.TEXT_PRIMARY};
                    font-weight: 500;
                }}
                QPushButton:hover:!checked {{
                    background: {Colors.BG_HOVER};
                }}
            """
            )
            btn.clicked.connect(lambda _, idx=i: self._select(idx))
            layout.addWidget(btn)
            self._buttons.append(btn)

    def _select(self, index):
        for i, btn in enumerate(self._buttons):
            btn.setChecked(i == index)
        self._selected = index
        self.selectionChanged.emit(self._buttons[index].text())

    def currentText(self):
        return self._buttons[self._selected].text()

    def currentIndex(self):
        return self._selected



from qtpy.QtCore import Signal
from qtpy.QtWidgets import QWidget, QHBoxLayout, QLineEdit, QLabel, QFrame
import qtawesome as qta

from ..stylesheets import Colors


class SearchWidget(QWidget):
    """
    Search widget with icon and clear button. Emits
    searchTextChanged signal when text changes.
    """

    searchTextChanged = Signal(str)

    def __init__(self, placeholder="Search...", parent=None):
        super().__init__(parent)
        self.setup_ui(placeholder)

    def setup_ui(self, placeholder):
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        container = QFrame()
        container.setStyleSheet(
            """
            QFrame {
                border: 1px solid #d1d5db;
                border-radius: 4px;
                background-color: transparent;
            }
            QFrame:focus-within {
                border: 1px solid #4f46e5;
            }
        """
        )
        container_layout = QHBoxLayout(container)
        container_layout.setContentsMargins(6, 0, 0, 0)
        container_layout.setSpacing(4)

        icon_label = QLabel()
        icon_label.setPixmap(
            qta.icon("ph.magnifying-glass", color=Colors.ICON).pixmap(16, 16)
        )
        icon_label.setFixedSize(16, 16)
        icon_label.setStyleSheet("border: none;")

        self.search_input = QLineEdit()
        self.search_input.setPlaceholderText(placeholder)
        self.search_input.setClearButtonEnabled(True)
        self.search_input.textChanged.connect(self.searchTextChanged.emit)
        self.search_input.setStyleSheet(
            """
            QLineEdit {
                border: none;
                padding: 6px 4px;
                background-color: transparent;
            }
        """
        )

        container_layout.addWidget(icon_label)
        container_layout.addWidget(self.search_input)
        layout.addWidget(container)

    def text(self):
        """Get current search text."""
        return self.search_input.text()

    def clear(self):
        """Clear search text."""
        self.search_input.clear()

    def setFocus(self):
        """Set focus to search input."""
        self.search_input.setFocus()



from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QFormLayout,
    QGroupBox,
    QComboBox,
    QWidget,
    QScrollArea,
    QTabWidget,
    QHBoxLayout,
    QPushButton,
    QFrame,
)
from qtpy.QtCore import Signal, QThread
import qtawesome as qta

from mosaic.settings import Settings
from mosaic.actor import QUALITY_PRESETS
from mosaic.stylesheets import (
    QPushButton_style,
    QScrollArea_style,
    QTabBar_style,
    Colors,
)

from mosaic.widgets import create_setting_widget, ColorPickerRow


class AppSettingsDialog(QDialog):

    settingsChanged = Signal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.settings = Settings.rendering
        self.quality_kwargs = Settings.vtk.get_settings()

        self.setWindowTitle("Settings")

        self.layout = QVBoxLayout(self)
        self.resize(400, 580)

        self.tabs = QTabWidget()
        self.build_tabs()

        self.layout.addWidget(self.tabs)

        from ..icons import dialog_accept_icon, dialog_reject_icon

        button_frame = QFrame()
        button_layout = QHBoxLayout(button_frame)
        button_layout.setContentsMargins(0, 10, 0, 0)
        reset_btn = QPushButton("Reset")
        reset_btn.setIcon(qta.icon("ph.arrow-counter-clockwise", color=Colors.ICON))
        reset_btn.clicked.connect(self.reset_settings)
        button_layout.addWidget(reset_btn)
        button_layout.addStretch()

        reject_btn = QPushButton("Cancel")
        reject_btn.setIcon(dialog_reject_icon)
        reject_btn.clicked.connect(super().reject)

        accept_btn = QPushButton("Done")
        accept_btn.setIcon(dialog_accept_icon)
        accept_btn.clicked.connect(super().accept)
        button_layout.addWidget(reject_btn)
        button_layout.addWidget(accept_btn)

        self.layout.addWidget(button_frame)
        self.setStyleSheet(QTabBar_style + QPushButton_style + QScrollArea_style)

    def build_tabs(self):
        self.tabs.clear()

        self.tabs.addTab(
            self.setup_general_page(),
            qta.icon("ph.gear", color=Colors.PRIMARY),
            "General",
        )
        self.tabs.addTab(
            self.setup_rendering_page(),
            qta.icon("ph.paint-brush", color=Colors.PRIMARY),
            "Data",
        )
        self.connect_signals()

        # Correctly initialize quality presets
        self.on_preset_changed()

        # Propagate defaults to render window
        self.settingsChanged.emit()

    def setup_general_page(self):
        """General settings page with colors, quality presets, and target FPS."""
        rendering = Settings.rendering

        page = QWidget()
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFrameShape(QScrollArea.Shape.NoFrame)

        scroll_widget = QWidget()
        layout = QVBoxLayout(scroll_widget)

        # Background color presets - dark and light theme options
        dark_presets = [
            (0.09, 0.10, 0.12),  # Cool slate (default)
            (0.00, 0.00, 0.00),  # Pure black
            (0.08, 0.12, 0.16),  # Deep ocean blue
            (0.12, 0.08, 0.14),  # Dark purple
            (0.06, 0.12, 0.10),  # Forest night
            (0.14, 0.10, 0.08),  # Dark bronze
        ]
        light_presets = [
            (0.97, 0.97, 0.96),  # Warm off-white (default)
            (1.00, 1.00, 1.00),  # Pure white
            (0.92, 0.95, 0.98),  # Ice blue
            (0.96, 0.94, 0.98),  # Lavender mist
            (0.94, 0.97, 0.94),  # Mint cream
            (0.98, 0.96, 0.92),  # Warm sand
        ]

        colors_group = QGroupBox("Background Colors")
        colors_layout = QVBoxLayout()
        colors_layout.setSpacing(16)

        self.bg_color_picker = ColorPickerRow(
            "Dark Background",
            default_color=rendering.background_color,
            preset_colors=dark_presets,
        )
        colors_layout.addWidget(self.bg_color_picker)

        self.bg_color_alt_picker = ColorPickerRow(
            "Light Background",
            default_color=rendering.background_color_alt,
            preset_colors=light_presets,
        )
        colors_layout.addWidget(self.bg_color_alt_picker)

        colors_group.setLayout(colors_layout)
        layout.addWidget(colors_group)

        perf_group = QGroupBox("Performance")
        perf_layout = QFormLayout()

        self.target_fps_spin = create_setting_widget(
            {
                "type": "float",
                "min": 1.0,
                "max": 144.0,
                "step": 1.0,
                "default": rendering.target_fps,
            }
        )

        self.parallel_worker_spin = create_setting_widget(
            {
                "type": "number",
                "min": 1,
                "max": QThread.idealThreadCount(),
                "step": 1,
                "default": rendering.parallel_worker,
            }
        )

        perf_layout.addRow("Target Frame Rate", self.target_fps_spin)
        perf_layout.addRow("Parallel Worker", self.parallel_worker_spin)

        perf_group.setLayout(perf_layout)
        layout.addWidget(perf_group)

        # Quality Preset Group (combined with parameters)
        quality_group = QGroupBox("Point Rendering")
        quality_layout = QFormLayout()

        cur_index, self.preset_combo = 0, QComboBox()
        for index, (preset_name, config) in enumerate(QUALITY_PRESETS.items()):
            description = config.get("description", "")
            display_text = f"{preset_name.title()}"
            if description:
                display_text += f" - {description}"
            self.preset_combo.addItem(display_text, preset_name)
            if preset_name == Settings.vtk.preset:
                cur_index = index

        self.preset_combo.setCurrentIndex(cur_index)
        quality_layout.addRow("Preset", self.preset_combo)

        # LOD Parameters (shown conditionally within the same group)
        self.lod_points_spin = create_setting_widget(
            {"type": "number", "min": 100000, "max": 50000000, "default": 5000000}
        )
        self.lod_points_row = quality_layout.rowCount()
        quality_layout.addRow("LOD Points", self.lod_points_spin)

        self.lod_point_size_spin = create_setting_widget(
            {"type": "number", "min": 1, "max": 20, "default": 3}
        )
        self.lod_point_size_row = quality_layout.rowCount()
        quality_layout.addRow("Point Size", self.lod_point_size_spin)

        quality_group.setLayout(quality_layout)
        self.quality_group = quality_group
        layout.addWidget(quality_group)

        layout.addStretch()
        scroll_area.setWidget(scroll_widget)

        page_layout = QVBoxLayout(page)
        page_layout.setContentsMargins(0, 0, 0, 0)
        page_layout.addWidget(scroll_area)

        return page

    def setup_rendering_page(self):
        """Rendering settings page."""
        page = QWidget()
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFrameShape(QScrollArea.Shape.NoFrame)
        scroll_widget = QWidget()
        layout = QVBoxLayout(scroll_widget)

        rendering = Settings.rendering

        # Anti-aliasing Group
        aa_group = QGroupBox("Anti-aliasing")
        aa_layout = QFormLayout()

        self.fxaa_check = create_setting_widget(
            {"type": "boolean", "default": rendering.enable_fxaa}
        )
        aa_layout.addRow("Enable FXAA", self.fxaa_check)

        self.multisamples_spin = create_setting_widget(
            {
                "type": "number",
                "min": 0,
                "max": 16,
                "default": rendering.multisamples,
            }
        )
        aa_layout.addRow("Multisamples", self.multisamples_spin)

        aa_group.setLayout(aa_layout)
        layout.addWidget(aa_group)

        # Smoothing Group
        smoothing_group = QGroupBox("Smoothing")
        smoothing_layout = QFormLayout()

        self.point_smooth_check = create_setting_widget(
            {"type": "boolean", "default": rendering.point_smoothing}
        )
        smoothing_layout.addRow("Point Smoothing", self.point_smooth_check)

        self.line_smooth_check = create_setting_widget(
            {"type": "boolean", "default": rendering.line_smoothing}
        )
        smoothing_layout.addRow("Line Smoothing", self.line_smooth_check)

        self.polygon_smooth_check = create_setting_widget(
            {"type": "boolean", "default": rendering.polygon_smoothing}
        )
        smoothing_layout.addRow("Polygon Smoothing", self.polygon_smooth_check)

        smoothing_group.setLayout(smoothing_layout)
        layout.addWidget(smoothing_group)

        # Depth Peeling Group
        depth_group = QGroupBox("Depth Peeling")
        depth_layout = QFormLayout()

        self.depth_peeling_check = create_setting_widget(
            {"type": "boolean", "default": rendering.use_depth_peeling}
        )
        depth_layout.addRow("Use Depth Peeling", self.depth_peeling_check)

        self.max_peels_spin = create_setting_widget(
            {
                "type": "number",
                "min": 1,
                "max": 20,
                "default": rendering.max_depth_peels,
            }
        )
        depth_layout.addRow("Max Depth Peels", self.max_peels_spin)

        self.occlusion_spin = create_setting_widget(
            {
                "type": "float",
                "min": 0.0,
                "max": 1.0,
                "step": 0.01,
                "default": rendering.occlusion_ratio,
            }
        )
        depth_layout.addRow("Occlusion Ratio", self.occlusion_spin)

        depth_group.setLayout(depth_layout)
        layout.addWidget(depth_group)

        layout.addStretch()
        scroll_area.setWidget(scroll_widget)

        # Set up the page layout
        page_layout = QVBoxLayout(page)
        page_layout.setContentsMargins(0, 0, 0, 0)
        page_layout.addWidget(scroll_area)

        return page

    def reset_settings(self):
        Settings.reset_to_defaults("vtk")
        Settings.reset_to_defaults("rendering")
        self.build_tabs()

    def connect_signals(self):
        """Connect all widget signals to auto-save to Settings and emit change."""

        def _update_setting(clx, attribute_name: str, value):
            setattr(clx, attribute_name, value)
            self.emit_parameters()

        self.widget_settings_map = {
            # vtk rendering settings
            self.target_fps_spin: (Settings.rendering, "target_fps", float),
            self.parallel_worker_spin: (Settings.rendering, "parallel_worker", int),
            self.fxaa_check: (Settings.rendering, "enable_fxaa", bool),
            self.multisamples_spin: (Settings.rendering, "multisamples", int),
            self.point_smooth_check: (Settings.rendering, "point_smoothing", bool),
            self.line_smooth_check: (Settings.rendering, "line_smoothing", bool),
            self.polygon_smooth_check: (Settings.rendering, "polygon_smoothing", bool),
            self.depth_peeling_check: (Settings.rendering, "use_depth_peeling", bool),
            self.max_peels_spin: (Settings.rendering, "max_depth_peels", int),
            self.occlusion_spin: (Settings.rendering, "occlusion_ratio", float),
            # vtk actor settings
            self.lod_points_spin: (Settings.vtk, "lod_points", int),
            self.lod_point_size_spin: (Settings.vtk, "lod_points_size", int),
        }

        for widget, (
            clx,
            attr_name,
            type_converter,
        ) in self.widget_settings_map.items():
            if hasattr(widget, "valueChanged"):
                widget.valueChanged.connect(
                    lambda v, clx=clx, attr=attr_name, conv=type_converter: _update_setting(
                        clx, attr, conv(v)
                    )
                )
            elif hasattr(widget, "toggled"):
                widget.toggled.connect(
                    lambda v, clx=clx, attr=attr_name: _update_setting(clx, attr, v)
                )

        self.preset_combo.currentTextChanged.connect(self.on_preset_changed)
        self.bg_color_picker.colorChanged.connect(
            lambda color: _update_setting(
                Settings.rendering,
                "background_color",
                color,
            )
        )
        self.bg_color_alt_picker.colorChanged.connect(
            lambda color: _update_setting(
                Settings.rendering,
                "background_color_alt",
                color,
            )
        )

    def on_preset_changed(self):
        """Handle preset selection change."""
        preset_name = self.preset_combo.currentData()
        if not preset_name:
            return None

        Settings.vtk.preset = preset_name
        preset_config = QUALITY_PRESETS.get(preset_name, {})
        quality_layout = self.quality_group.layout()

        self.lod_points_spin.blockSignals(True)
        self.lod_point_size_spin.blockSignals(True)

        if "lod_points" in preset_config:
            self.lod_points_spin.setValue(int(preset_config["lod_points"]))
        if "lod_points_size" in preset_config:
            self.lod_point_size_spin.setValue(preset_config["lod_points_size"])

        self.lod_points_spin.blockSignals(False)
        self.lod_point_size_spin.blockSignals(False)

        quality_type = preset_config.get("quality", "full")
        Settings.vtk.quality = quality_type

        lod_visible = quality_type == "lod"
        quality_layout.setRowVisible(self.lod_points_row, lod_visible)
        quality_layout.setRowVisible(self.lod_point_size_row, lod_visible)

        return self.emit_parameters()

    def emit_parameters(self):
        """Emit parameters change signal (match GeometryPropertiesDialog pattern)."""
        self.settingsChanged.emit()



#!python3
"""
Pipeline entrypoint.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import sys
import json
import argparse
from pathlib import Path
from concurrent.futures import ProcessPoolExecutor, as_completed

from mosaic.pipeline.executor import generate_runs, execute_run


def run_wrapper(run_config, skip_complete: bool = False):

    run_id = run_config["run_id"]
    try:
        execute_run(run_config, skip_complete)
        return run_id, None
    except Exception as e:
        return run_id, str(e)


def main():
    parser = argparse.ArgumentParser(
        description="Execute Mosaic pipelines from the command line",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  mosaic-pipeline config.json
  mosaic-pipeline config.json --workers 8
  mosaic-pipeline config.json --index 0
  mosaic-pipeline config.json --index $SLURM_ARRAY_TASK_ID
        """,
    )
    parser.add_argument("config", type=Path, help="Pipeline configuration JSON file")
    parser.add_argument(
        "-w",
        "--workers",
        type=int,
        default=1,
        help="Number of parallel workers (default: 1)",
    )
    parser.add_argument(
        "-i",
        "--index",
        type=int,
        default=None,
        help="Run specific index (for job arrays)",
    )
    parser.add_argument(
        "--skip-complete", action="store_true", help="Do not run completed jobs again"
    )
    parser.add_argument(
        "-n", "--dry-run", action="store_true", help="List runs without executing"
    )

    args = parser.parse_args()

    if not args.config.exists():
        print(f"Error: {args.config} not found", file=sys.stderr)
        return 1

    with open(args.config, mode="r") as ifile:
        pipeline_config = json.load(ifile)

    try:
        runs = generate_runs(pipeline_config)
    except Exception as e:
        print(f"Error generating runs: {e}", file=sys.stderr)
        return 1

    if not runs:
        print("No runs to execute", file=sys.stderr)
        return 1

    if args.dry_run:
        print(f"Total runs: {len(runs)}")
        for i, run in enumerate(runs):
            print(f"  [{i}] {run['run_id']}")
        return 0

    if args.index is not None:
        if args.index < 0 or args.index >= len(runs):
            print(
                f"Error: index {args.index} out of range [0, {len(runs)-1}]",
                file=sys.stderr,
            )
            return 1
        runs = [runs[args.index]]
        print(f"Executing run {args.index}: {runs[0]['run_id']}")

    total = len(runs)

    completed = 0
    failed = 0
    with ProcessPoolExecutor(max_workers=args.workers, max_tasks_per_child=1) as pool:
        futures = {
            pool.submit(run_wrapper, run, args.skip_complete): run for run in runs
        }

        for future in as_completed(futures):
            run = futures[future]
            completed += 1
            try:
                run_id, error = future.result()
                if error:
                    failed += 1
                    print(
                        f"[{completed}/{total}] {run_id}: Error - {error}",
                        file=sys.stderr,
                    )
                else:
                    print(f"[{completed}/{total}] {run_id}: Done")
            except Exception as e:
                failed += 1
                print(
                    f"[{completed}/{total}] {run['run_id']}: "
                    f"Unexpected error - {e}",
                    file=sys.stderr,
                )

    if failed:
        print(f"{failed}/{total} runs failed", file=sys.stderr)
    return 1 if failed else 0


if __name__ == "__main__":
    sys.exit(main())



from .path_selector import PathSelector
from .dialog_footer import DialogFooter
from .search_widget import SearchWidget
from .sidebar import ObjectBrowserSidebar
from .sliders import DualHandleSlider, SliderRow, HistogramRangeSlider
from .trajectory_player import TrajectoryPlayer
from .ribbon import RibbonToolBar, create_button
from .colors import (
    ColorSwatch,
    ColorPickerRow,
    ColorMapSelector,
    ColorMapSelector,
    generate_gradient_colors,
)
from .volume_viewer import MultiVolumeViewer, VolumeViewer
from .vtk_widgets import (
    AxesWidget,
    BoundingBoxManager,
    LegendWidget,
    ScaleBarWidget,
)
from .container_list import (
    ContainerListWidget,
    ContainerTreeWidget,
    GroupTreeWidgetItem,
    StyledListWidgetItem,
    StyledTreeWidgetItem,
)
from .status_indicator import StatusIndicator, CursorModeHandler, ViewerModes
from .settings import *



"""
Implements ProgressDialog, showing progress in the status bar.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from qtpy.QtWidgets import QMessageBox


class ProgressDialog:
    def __init__(self, iterable, title="Processing", parent=None):
        from ..widgets.status_indicator import StatusIndicator

        self.total = len(iterable)
        self.iterator = iter(iterable)
        self.current = 0
        self._indicator = StatusIndicator.instance()

        if self._indicator is not None:
            self._indicator.show_progress(title, self.total)

    def update_progress(self, value: int = None):
        """Update progress bar and status label."""
        if value is None:
            value = self.current

        if self._indicator is not None:
            self._indicator.update_progress(value, self.total)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = next(self.iterator)
            self.update_progress()
            self.current += 1
            return item
        except StopIteration:
            self.update_progress()
            self.close()
            raise

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()
        if exc_type is not None:
            QMessageBox.warning(None, "Error", str(exc_value))
            return True

    def close(self):
        """Hide the status bar progress."""
        if self._indicator is not None:
            self._indicator.hide_progress()



from functools import partial

import numpy as np
from qtpy.QtWidgets import QWidget, QVBoxLayout

from .. import meshing
from .. import operations as _operations  # noqa: F401  # registers geometry operations
from ..registry import MethodRegistry
from ..widgets.ribbon import create_button
from ..parallel import submit_task, submit_task_batch


def _repair_mesh(
    geometry,
    max_hole_size=-1,
    smoothness=0,
    curvature_weight=0,
    pressure=0,
    flip_normals=False,
    fair_all=False,
    boundary_ring=0,
):
    import igl
    from .. import meshing
    from ..parametrization import TriangularMesh

    fair = not (smoothness == 0 and curvature_weight == 0 and pressure == 0)

    model = geometry.model
    model.mesh.remove_non_manifold_edges()
    model.mesh.remove_degenerate_triangles()
    model.mesh.remove_duplicated_triangles()
    model.mesh.remove_unreferenced_vertices()
    model.mesh.remove_duplicated_vertices()

    vs = model.vertices
    fs = model.triangles

    out_fs = meshing.close_holes(vs, fs, max_hole_size)
    if fair:
        hole_fids = np.arange(len(fs), len(out_fs))

        try:
            mesh = meshing.remesh(meshing.to_open3d(vs, out_fs))
            new_vs = np.asarray(mesh.vertices, dtype=np.float64)
            fs = np.asarray(mesh.triangles)
        except (ValueError, RuntimeError):
            new_vs, fs = vs, out_fs

        if fair_all:
            vids = np.arange(len(new_vs))
        else:
            _, face_ids, _ = igl.point_mesh_squared_distance(
                new_vs, vs, out_fs.astype(np.int64)
            )
            vids = np.where(np.isin(face_ids, hole_fids))[0]

        vs = new_vs
        if len(vids) > 0:
            vs = meshing.fair_mesh(
                vs,
                fs,
                vids,
                smoothness=smoothness,
                curvature_weight=curvature_weight,
                pressure=pressure,
                n_ring=boundary_ring,
            )

    if flip_normals:
        fs = fs[:, ::-1]

    from ..geometry import GeometryData

    return GeometryData(
        model=TriangularMesh(meshing.to_open3d(vs, fs)),
        sampling_rate=geometry.sampling_rate.copy(),
        meta=geometry._meta.copy(),
    )


def _fill_mesh(mesh_geometry):
    from ..geometry import GeometryData

    model = mesh_geometry.model
    voxel_size = max(mesh_geometry.sampling_rate)
    points = meshing.fill_mesh(model.vertices, model.triangles, voxel_size=voxel_size)
    return GeometryData(points=points, sampling_rate=mesh_geometry.sampling_rate.copy())


def _project(
    mesh_geometries,
    geometries,
    use_normals: bool = False,
    invert_normals: bool = False,
    update_normals: bool = False,
    partition: bool = False,
):
    from ..geometry import Geometry, GeometryData

    meshes = [mg.model for mg in mesh_geometries]
    n_meshes = len(meshes)

    data_out, meshes_out = [], []
    mesh_subsets = [[] for _ in range(n_meshes)]
    mesh_proj = [[] for _ in range(n_meshes)]
    mesh_tri = [[] for _ in range(n_meshes)]

    for geometry in geometries:
        normals = geometry.normals if use_normals else None
        if normals is not None:
            normals = normals * (-1 if invert_normals else 1)

        all_dist, all_proj, all_tri = [], [], []
        for mesh in meshes:
            dist, proj, tri = mesh.compute_distance(
                points=geometry.points,
                normals=normals,
                return_projection=True,
                return_indices=False,
                return_triangles=True,
            )
            all_dist.append(dist)
            all_proj.append(proj)
            all_tri.append(tri)

        best = np.argmin(np.stack(all_dist), axis=0)
        proj_sel = np.empty_like(all_proj[0])

        for m in range(n_meshes):
            mask = best == m
            if not mask.any():
                continue
            if partition:
                mesh_subsets[m].append(geometry[mask])
            else:
                proj_m = all_proj[m][mask]
                proj_sel[mask] = proj_m
                mesh_proj[m].append(proj_m)
                mesh_tri[m].append(all_tri[m][mask])

        if not partition:
            geo_normals = geometry.normals
            if update_normals:
                geo_normals = np.empty((len(geometry.points), 3))
                for m in range(n_meshes):
                    mask = best == m
                    if mask.any():
                        geo_normals[mask] = meshes[m].compute_normal(proj_sel[mask])
            data_out.append(
                GeometryData(
                    points=proj_sel,
                    normals=geo_normals,
                    sampling_rate=geometry.sampling_rate.copy(),
                )
            )

    if partition:
        for m, subsets in enumerate(mesh_subsets):
            if subsets:
                geom = subsets[0] if len(subsets) == 1 else Geometry.merge(subsets)
                name = mesh_geometries[m]._meta.get("name", f"Mesh {m}")
                geom._meta["name"] = f"{name}_partition"
                data_out.append(geom)
    else:
        for m in range(n_meshes):
            if not mesh_proj[m]:
                continue
            new_model = meshes[m].add_projections(
                np.concatenate(mesh_proj[m]),
                np.concatenate(mesh_tri[m]),
                return_indices=False,
            )
            meshes_out.append(
                GeometryData(
                    model=new_model,
                    sampling_rate=mesh_geometries[m].sampling_rate.copy(),
                )
            )

    return data_out + meshes_out


class ModelTab(QWidget):
    def __init__(self, cdata, ribbon, legend, **kwargs):
        super().__init__()
        self.cdata = cdata
        self.ribbon = ribbon
        self.legend = legend

        layout = QVBoxLayout(self)
        layout.setSpacing(5)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.ribbon)

    def show_ribbon(self):
        self.ribbon.clear()

        func = self._fit_parallel
        fitting_actions = [
            create_button(
                "Sphere",
                "ph.circle",
                self,
                partial(func, "sphere"),
                "Fit to selected clusters",
            ),
            create_button(
                "Ellipse",
                "ph.link-simple-horizontal-break",
                self,
                partial(func, "ellipsoid"),
                "Fit to selected clusters",
            ),
            create_button(
                "Cylinder",
                "ph.hexagon",
                self,
                partial(func, "cylinder"),
                "Fit to selected clusters",
            ),
            create_button(
                "RBF",
                "ph.dots-nine",
                self,
                partial(func, "rbf"),
                "Fit to selected clusters",
                RBF_SETTINGS,
            ),
            create_button(
                "Mesh",
                "ph.triangle",
                self,
                func,
                "Fit to selected clusters",
                MethodRegistry.settings_dict("fit"),
            ),
            create_button(
                "Curve",
                "ph.line-segments",
                self,
                partial(func, "spline"),
                "Fit to selected clusters",
                SPLINE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Parametrization", fitting_actions)

        mesh_actions = [
            create_button(
                "Sample",
                "ph.broadcast",
                self,
                self._sample_parallel,
                "Generate points from fitted model",
                SAMPLE_SETTINGS,
            ),
        ]
        self.ribbon.add_section("Sampling", mesh_actions)

        mesh_actions = [
            create_button(
                "Repair",
                "ph.wrench",
                self,
                self._repair_mesh_parallel,
                "Fix holes and topology issues",
                REPAIR_SETTINGS,
            ),
            create_button(
                "Remesh",
                "ph.arrows-clockwise",
                self,
                self._remesh_parallel,
                "Adjust resolution and quality",
                MethodRegistry.settings_dict("remesh"),
            ),
            create_button(
                "Smooth",
                "ph.drop",
                self,
                self._smooth_parallel,
                "Reduce surface noise",
                MethodRegistry.settings_dict("smooth"),
            ),
            create_button(
                "Project",
                "ph.arrow-line-down",
                self,
                self._project_parallel,
                "Project points onto mesh",
                PROJECTION_SETTINGS,
            ),
            create_button(
                "Fill",
                "ph.cube",
                self,
                self._fill_parallel,
                "Fill the interior of a closed mesh with points",
            ),
        ]
        self.ribbon.add_section("Mesh Operations", mesh_actions)

    def _default_callback(self, geom):
        from ..geometry import Geometry, GeometryData
        from ..parametrization import TriangularMesh

        if isinstance(geom, (Geometry, GeometryData)):
            geom = (geom,)

        new_model, new_cluster = False, False
        for new_geom in geom:

            if isinstance(new_geom, GeometryData):
                new_geom = Geometry(**new_geom.to_dict())

            if isinstance(new_geom.model, TriangularMesh):
                new_geom.change_representation("surface")

            if new_geom.model is None:
                new_cluster = True
                self.cdata.data.add(new_geom)
                continue

            new_model = True
            self.cdata.models.add(new_geom)

        if new_model:
            self.cdata.models.render()
        if new_cluster:
            self.cdata.data.render()

    def _get_selected_meshes(self):
        from ..parametrization import TriangularMesh

        ret = []
        for geometry in self.cdata.models.get_selected_geometries():
            fit = geometry.model
            if not isinstance(fit, TriangularMesh):
                continue
            ret.append(geometry)
        return ret

    def _repair_mesh_parallel(self, **kwargs):
        for geometry in self._get_selected_meshes():
            submit_task(
                "Repair Mesh",
                _repair_mesh,
                self._default_callback,
                geometry._geometry_data,
                **kwargs,
            )

    def _fit_parallel(self, method: str, *args, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self.cdata.data.get_selected_geometries():
            submit_task(
                "Parametrization",
                GeometryOperations.fit,
                self._default_callback,
                geometry._geometry_data,
                method,
                **kwargs,
            )

    def _smooth_parallel(self, method, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self._get_selected_meshes():
            submit_task(
                "Smooth",
                GeometryOperations.smooth,
                self._default_callback,
                geometry._geometry_data,
                method,
                **kwargs,
            )

    def _sample_parallel(self, sampling, method, normal_offset=0.0, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self.cdata.models.get_selected_geometries():
            submit_task(
                "Sample Fit",
                GeometryOperations.sample,
                self._default_callback,
                geometry._geometry_data,
                method=method,
                sampling=sampling,
                normal_offset=normal_offset,
                **kwargs,
            )

    def _remesh_parallel(self, method, **kwargs):
        from ..operations import GeometryOperations

        for geometry in self._get_selected_meshes():
            submit_task(
                "Remesh",
                GeometryOperations.remesh,
                self._default_callback,
                geometry._geometry_data,
                method,
                **kwargs,
            )

    def _fill_parallel(self, **kwargs):
        for geometry in self._get_selected_meshes():
            submit_task(
                "Fill Mesh", _fill_mesh, self._default_callback, geometry._geometry_data
            )

    def _project_parallel(
        self,
        use_normals: bool = False,
        invert_normals: bool = False,
        update_normals: bool = False,
        partition: bool = False,
        **kwargs,
    ):
        selected_meshes = self._get_selected_meshes()
        if not selected_meshes:
            raise ValueError("Please select at least one mesh for projection.")

        submit_task(
            "Project",
            _project,
            self._default_callback,
            selected_meshes,
            self.cdata.data.get_selected_geometries(),
            use_normals,
            invert_normals,
            update_normals,
            partition,
        )


SAMPLE_SETTINGS = MethodRegistry.settings_dict("sample")

RBF_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Direction",
            "parameter": "direction",
            "type": "select",
            "options": ["xy", "xz", "yz"],
            "default": "xy",
            "description": "Coordinate plane to fit RBF in.",
        },
    ],
}

SPLINE_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Order",
            "parameter": "order",
            "type": "number",
            "default": 3,
            "min": 1,
            "max": 5,
            "description": "Spline order to fit to control points.",
        },
    ],
}

REPAIR_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Smoothness",
            "parameter": "smoothness",
            "type": "float",
            "default": 0.0,
            "min": 0.0,
            "max": 1.0,
            "description": "Balance between position anchoring and curvature "
            "minimization. 0 = stay in place, 1 = full smoothing.",
        },
        {
            "label": "Curvature Weight",
            "parameter": "curvature_weight",
            "type": "float",
            "default": 0.0,
            "description": "Higher-order smoothing for curvature continuity.",
        },
        {
            "label": "Pressure",
            "parameter": "pressure",
            "type": "float",
            "default": 0.0,
            "description": "Internal mesh pressure along vertex normals.",
        },
        {
            "label": "Hole Size",
            "parameter": "max_hole_size",
            "type": "float",
            "min": -1.0,
            "default": -1.0,
            "description": "Maximum surface area of holes considered for triangulation.",
        },
        {
            "label": "Flip Normals",
            "parameter": "flip_normals",
            "type": "boolean",
            "default": False,
            "description": "Reverse normal direction of the mesh.",
        },
        {
            "label": "Fair All Vertices",
            "parameter": "fair_all",
            "type": "boolean",
            "default": False,
            "description": "Apply fairing to all vertices, not just inferred ones.",
        },
        {
            "label": "Boundary Ring",
            "parameter": "boundary_ring",
            "type": "number",
            "default": 0,
            "min": 0,
            "description": "Number of vertex rings around inferred vertices to include in fairing.",
        },
    ],
}


PROJECTION_SETTINGS = {
    "title": "Settings",
    "settings": [
        {
            "label": "Cast Normals",
            "parameter": "use_normals",
            "type": "boolean",
            "default": True,
            "description": "Include normal vectors in raycasting.",
        },
        {
            "label": "Invert Normals",
            "parameter": "invert_normals",
            "type": "boolean",
            "default": False,
            "description": "Invert direction of normal vectors.",
        },
        {
            "label": "Update Normals",
            "parameter": "update_normals",
            "type": "boolean",
            "default": False,
            "description": "Update normal vectors of projection based on the mesh.",
        },
        {
            "label": "Partition",
            "parameter": "partition",
            "type": "boolean",
            "default": False,
            "description": "Assign points to nearest mesh instead of projecting.",
        },
    ],
}



"""
Implements MosaicData, which is reponsible for tracking overall
application state and mediating interaction between segmentations
and parametrizations.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

__all__ = ["MosaicData"]


class MosaicData:
    """
    Initialize MosaicData instance for managing application state.

    Parameters
    ----------
    vtk_widget : VTKWidget
        VTK widget instance for 3D visualization
    """

    def __init__(self, vtk_widget):
        super().__init__()
        from .container import DataContainer
        from .interactor import DataContainerInteractor

        # Data containers and GUI interaction elements
        self.shape = None
        self._data = DataContainer()
        self._models = DataContainer(highlight_color=(0.2, 0.4, 0.8))

        self.data = DataContainerInteractor(self._data, vtk_widget)
        self.models = DataContainerInteractor(self._models, vtk_widget, prefix="Fit")

        self.data.attach_area_picker()
        self.active_picker = "data"

    def to_file(self, filename: str):
        """Save current application state to file.

        Parameters
        ----------
        filename : str
            Path to save the application state.
        """
        import pickle

        state = {
            "shape": self.shape,
            "_data": self._data,
            "_models": self._models,
            "_data_tree": self.data.data_list.to_state(),
            "_models_tree": self.models.data_list.to_state(),
        }
        with open(filename, "wb") as ofile:
            pickle.dump(state, ofile)

    def load_session(self, filename: str):
        """
        Load application state from file.

        Parameters
        ----------
        filename : str
            Path to the saved session file (.pickle).
        """
        from .container import DataContainer
        from .formats import open_file, open_session

        sampling = 1
        cluster_tree, model_tree = None, None
        if filename.endswith("pickle"):
            data = open_session(filename)
            shape = data["shape"]
            cluster_container, model_container = data["_data"], data["_models"]
            cluster_tree = data.get("_data_tree")
            model_tree = data.get("_models_tree")

        else:
            container = open_file(filename)

            shape = container.shape
            sampling = container.sampling
            cluster_container, model_container = DataContainer(), DataContainer()
            for data in container:
                cluster_container.add(
                    points=data.vertices, normals=data.normals, sampling_rate=sampling
                )

        metadata = {"shape": self.shape, "sampling_rate": sampling}

        cluster_container.metadata = metadata.copy()
        model_container.metadata = metadata.copy()

        self.shape = shape
        self.data.update(cluster_container, tree_state=cluster_tree)
        self.models.update(model_container, tree_state=model_tree)

    def reset(self):
        """
        Reset the state of the class instance.
        """
        from .container import DataContainer

        self.shape = None
        self.data.update(DataContainer())
        self.models.update(DataContainer(highlight_color=(0.2, 0.4, 0.8)))

    def refresh_actors(self):
        """
        Reinitialize all vtk actors to accomodate render setting changes.
        """
        self.data.refresh_actors()
        self.models.refresh_actors()

    def set_coloring_mode(self, mode: str):
        self.data.set_coloring_mode(mode)
        self.models.set_coloring_mode(mode)

    def _get_active_container(self):
        if self.active_picker == "data":
            return self.data
        return self.models

    def swap_area_picker(self):
        """Toggle area picker between data and models containers."""
        self.active_picker = "data" if self.active_picker != "data" else "models"
        self.data.activate_viewing_mode()
        self.models.activate_viewing_mode()
        container = self._get_active_container()
        return container.attach_area_picker()

    def activate_viewing_mode(self):
        """Activate viewing mode for all contaienrs."""
        self.data.activate_viewing_mode()
        self.models.activate_viewing_mode()

    def highlight_clusters_from_selected_points(self):
        """Highlight clusters containing currently selected points.

        Returns
        -------
        bool
            Success status of highlighting operation
        """
        obj = self._get_active_container()
        return obj.highlight_clusters_from_selected_points()

    def visibility_unselected(self, visible: bool = True):
        """Hide clusters and models that are not selected."""
        cluster = list(self.data.point_selection.keys())
        cluster.extend(self.data._get_selected_uuids())
        cluster = set(cluster)

        unselected = self.data.data_list.to_state().get_all_uuids() - cluster
        self.data.visibility(
            geometries=[self._data.get(x) for x in unselected], visible=visible
        )

        models = set(self.models._get_selected_uuids())
        unselected = self.models.data_list.to_state().get_all_uuids() - models
        self.models.visibility(
            geometries=[self._models.get(x) for x in unselected], visible=visible
        )

    def activate_picking_mode(self):
        obj = self._get_active_container()
        return obj.activate_picking_mode()

    def format_datalist(
        self, type="data", mesh_only: bool = False, selected: bool = False
    ):
        """Format data list for dialog display.

        Parameters
        ----------
        type : str, optional
            Type of data to format ('data' or 'models'), by default 'data'
        mesh_only : bool, optional
            Whether to return only TriangularMesh instances for type 'models'.
        selected : bool, optional
            Whether to return only selected objects

        Returns
        -------
        list
            List of tuples containing (item_text, data_object) pairs
        """
        if mesh_only and type != "models":
            mesh_only = False

        interactor, container = self.data, self._data
        if type == "models":
            interactor, container = self.models, self._models

        selection = [x.uuid for x in container.data]
        if selected:
            selection = interactor._get_selected_uuids()

        ret = []
        for i in selection:
            if (geometry := container.get(i)) is None:
                continue

            if mesh_only:
                from .parametrization import TriangularMesh

                is_mesh = isinstance(geometry.model, TriangularMesh)
                if not is_mesh:
                    continue

            ret.append((geometry._meta.get("name", ""), geometry))
        return ret

    def get_tree_state(self, type="data"):
        """Get the tree state (group structure) for a container.

        Parameters
        ----------
        type : str, optional
            Type of container ('data' or 'models'), by default 'data'

        Returns
        -------
        TreeStateData
            The current tree structure including groups.
        """
        if type == "models":
            return self.models.data_list.to_state()
        return self.data.data_list.to_state()



"""
Mosaic command interface — headless command dispatch and interactive shell.

Copyright (c) 2026 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

from .session import Session
from .registry import CommandRegistry

__all__ = ["Session", "CommandRegistry"]



import qtawesome as qta

from .stylesheets import Colors

dialog_margin = (16, 16, 16, 16)
footer_margin = (0, 16, 0, 0)

dialog_accept_icon = qta.icon("ph.caret-right", color=Colors.PRIMARY)
dialog_reject_icon = qta.icon("ph.x", color=Colors.ICON)

dialog_next_icon = qta.icon("ph.skip-forward", color=Colors.PRIMARY)
dialog_previous_icon = qta.icon("ph.skip-back", color=Colors.PRIMARY)
dialog_apply_icon = qta.icon("ph.checks", color=Colors.PRIMARY)

dialog_selectall_icon = qta.icon("ph.check-square", color=Colors.PRIMARY)
dialog_selectnone_icon = qta.icon("ph.x-square", color=Colors.PRIMARY)

info_icon = qta.icon("ph.info", color=Colors.PRIMARY).pixmap(18, 18)

cluster_icon = None
model_icon = None



"""
DTS parameter screening: template generation, screen management, status.

Copyright (c) 2025 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import json
import shutil
import textwrap
import warnings
from pathlib import Path
from os import makedirs
from datetime import datetime
from itertools import product
from typing import Dict, List, Optional

import numpy as np

from ..parallel import report_progress

__all__ = [
    "generate_screen",
    "extend_screen",
    "get_screen_status",
    "run_status",
    "write_launcher_scripts",
]


def run_status(run_dir: Path) -> str:
    """Return ``"done"`` if energy xvg or .res files exist, else ``"pending"``."""
    if (run_dir / "dts-en.xvg").exists():
        return "done"
    return "done" if list(run_dir.glob("*.res")) else "pending"


def _prepare_volume(
    volume_path: str,
    output_dir: str,
    use_filters: bool = False,
    lowpass_cutoff: Optional[float] = None,
    highpass_cutoff: Optional[float] = None,
    plane_norm: Optional[str] = None,
) -> str:
    """Apply bandpass filtering to volume if requested.

    Parameters
    ----------
    volume_path : str
        Path to input volume.
    output_dir : str
        Directory for filtered output.
    use_filters : bool
        Whether to apply filtering.
    lowpass_cutoff : float
        Lowpass filter cutoff.
    highpass_cutoff : float
        Highpass filter cutoff.
    plane_norm : str
        Axis to normalize along (x/y/z or None).

    Returns
    -------
    str
        Path to (possibly filtered) volume.
    """
    if not use_filters:
        return volume_path

    from tme import Density

    try:
        from tme.filters import BandPassFilter
    except ImportError:
        from tme.filters import BandPassReconstructed as BandPassFilter

    data = Density.from_file(volume_path)
    sampling, origin = data.sampling_rate, data.origin

    bpf = BandPassFilter(
        lowpass=lowpass_cutoff,
        highpass=highpass_cutoff,
        sampling_rate=np.max(sampling),
        use_gaussian=True,
    )
    template_ft = np.fft.rfftn(data.data, s=data.shape)
    mask = bpf(shape=data.shape, return_real_fourier=True)["data"]
    template_ft = np.multiply(template_ft, mask, out=template_ft)
    filtered = np.fft.irfftn(template_ft, s=data.shape).real

    axis_map = {"x": 0, "y": 1, "z": 2}
    axis = axis_map.get(plane_norm, None)
    if axis is not None:
        axis = tuple(i for i in range(filtered.ndim) if i != axis)
        filtered = filtered / filtered.max(axis=axis, keepdims=True)

    volume_path = str(Path(output_dir) / "density.mrc")
    filtered = filtered.astype(np.float32)
    Density(filtered, origin=origin, sampling_rate=sampling).to_file(volume_path)
    return volume_path


def _build_dts_template(
    volume_path: Optional[str],
    mesh_scale: str,
    mesh_offset: str,
    hmff_params: Dict,
    sim_params: Dict,
    extra_config: str = "",
    coupling_params: Optional[Dict] = None,
) -> str:
    """Build a DTS config template with screening placeholders.

    Values that should be screened must be pre-formatted as
    ``{{name:range}}`` strings by the caller (dialog).

    Parameters
    ----------
    volume_path : str or None
        Path to volume file.  When *None* the standard Helfrich
        energy method is used instead of HMFF.
    mesh_scale : str
        Scale factor from mesh.txt.
    mesh_offset : str
        Offset from mesh.txt.
    hmff_params : dict
        Physical parameters: kappa, kappa0 and optionally xi,
        invert_contrast for HMFF.
        Screened values are already ``{{name:range}}`` strings.
    sim_params : dict
        Simulation values: steps, threads, temperature.
    extra_config : str
        Raw DTS config lines appended verbatim.  May contain
        ``{{name:range}}`` placeholders for additional screening.
    coupling_params : dict, optional
        Coupling configurations keyed by coupling name.  Each value
        is ``{"mode": str, "values": list}``.

    Returns
    -------
    str
        DTS config template content.
    """
    kappa = hmff_params.get("kappa", 25.0)
    kappa0 = hmff_params.get("kappa0", 0.0)

    steps = sim_params.get("steps", 50000)
    threads = sim_params.get("threads", 1)
    temperature = sim_params.get("temperature", 1.0)
    min_edge = sim_params.get("min_edge", 1.0)
    max_edge = sim_params.get("max_edge", 5.0)
    output_period = sim_params.get("output_period", 1000)

    integrator = "MetropolisAlgorithm"
    try:
        if int(threads) > 1:
            integrator = "MetropolisAlgorithmOpenMP"
    except (ValueError, TypeError):
        pass

    if volume_path is not None:
        xi = hmff_params.get("xi", 5.0)
        invert_contrast = hmff_params.get("invert_contrast", True)
        invert_flag = (
            int(invert_contrast)
            if isinstance(invert_contrast, bool)
            else invert_contrast
        )
        energy_line = (
            f"EnergyMethod             = FreeDTS1.0_MDFF {volume_path} {xi} 0 "
            f"{mesh_scale} {mesh_offset} {invert_flag} 0"
        )
    else:
        energy_line = "EnergyMethod             = FreeDTS1.0"

    coupling_params = coupling_params or {}
    coupling_lines = ""
    for ckey, dts_keyword in [
        ("vol_coupling", "VolumeCoupling"),
        ("curv_coupling", "GlobalCurvatureCoupling"),
        ("area_coupling", "TotalAreaCoupling"),
    ]:
        cfg = coupling_params.get(ckey)
        if not cfg:
            continue
        mode = cfg["mode"]
        values = " ".join(str(v) for v in cfg["values"])
        coupling_lines += f"{dts_keyword:<25s}= {mode} {values}\n"

    dts_config = textwrap.dedent(
        f"""\
        {energy_line}
        Integrator_Type          = MC_Simulation
        VertexPositionIntegrator = {integrator} 1 1 0.05
        AlexanderMove            = {integrator} 1
        InclusionPoseIntegrator  = MetropolisAlgorithm 1 1
        VisualizationFormat      = VTUFileFormat VTU_F {output_period}
        NonbinaryTrajectory      = TSI TrajTSI {output_period}
        Kappa                    = {kappa} {kappa0} {kappa0}
        Temperature              = {temperature} 0
        Set_Steps                = 1 {steps}
        Min_Max_Lenghts          = {min_edge} {max_edge}
        TimeSeriesData_Period    = {output_period}
        Box_Centering_F          = 0
        """
    )

    if coupling_lines:
        dts_config += coupling_lines

    if extra_config and extra_config.strip():
        dts_config += extra_config.strip() + "\n"

    return dts_config


def _expand_screen(
    screen_dir: Path,
    template_content: str,
    param_values: Dict[str, List],
    existing_runs: Optional[List[Dict]] = None,
) -> List[Dict]:
    """Generate new parameter combinations and create run directories.

    Parameters
    ----------
    screen_dir : Path
        Screen output directory.
    template_content : str
        Processed template with ``{param}`` placeholders.
    param_values : dict
        Parameter names mapped to value lists.
    existing_runs : list of dict, optional
        Already existing runs whose combinations are skipped.

    Returns
    -------
    list of dict
        Newly created run info dicts with ``run_id`` and ``parameters``.
    """
    existing_runs = existing_runs or []
    existing_combos = {tuple(sorted(r["parameters"].items())) for r in existing_runs}

    param_names = sorted(param_values.keys())
    all_values = [param_values[name] for name in param_names]

    new_runs = []
    run_counter = len(existing_runs)

    for combo in product(*all_values):
        params = dict(zip(param_names, combo))
        combo_key = tuple(sorted(params.items()))
        if combo_key in existing_combos:
            continue

        run_counter += 1
        width = max(len(str(run_counter)), 4)
        run_id = f"run_{run_counter:0{width}d}"
        run_dir = screen_dir / run_id
        run_dir.mkdir(exist_ok=True)

        content = template_content
        for pname, pval in params.items():
            content = content.replace(f"{{{pname}}}", str(pval))

        (run_dir / "input.dts").write_text(content, encoding="utf-8")
        with open(run_dir / "params.json", "w") as f:
            json.dump(params, f, indent=2)

        new_runs.append({"run_id": run_id, "parameters": params})

    return new_runs


def _write_summary(
    screen_dir: Path,
    template_file: str,
    parameters: Dict[str, List],
    runs: List[Dict],
) -> Dict:
    """Write ``screen_summary.json``.

    Parameters
    ----------
    screen_dir : Path
        Screen output directory.
    template_file : str
        Path to template file (stored in summary for reference).
    parameters : dict
        Parameter names mapped to value lists.
    runs : list of dict
        All runs (existing + new).

    Returns
    -------
    dict
        The written summary.
    """
    summary = {
        "template_file": template_file,
        "generated_at": datetime.now().isoformat(),
        "total_runs": len(runs),
        "parameters": sorted(parameters.keys()),
        "parameter_ranges": {
            name: {
                "values": values,
                "min": min(values),
                "max": max(values),
                "count": len(values),
            }
            for name, values in parameters.items()
        },
        "runs": [{"run_id": r["run_id"], "parameters": r["parameters"]} for r in runs],
    }
    with open(screen_dir / "screen_summary.json", "w") as f:
        json.dump(summary, f, indent=2)
    return summary


def _write_run_script(run_dir: Path, params: Dict) -> None:
    """Write a run.sh script for a single screen run.

    Parameters
    ----------
    run_dir : Path
        Directory for this run (must contain input.dts and topol.top).
    params : dict
        Run parameters (may contain 'steps' and 'threads' overrides).
    """
    cmd = "DTS"
    if shutil.which("dts"):
        cmd = "dts"

    topol = run_dir / "topol.top"
    mesh = ""
    if topol.exists():
        mesh = topol.read_text().split()[0]

    input_file = run_dir / "input.dts"
    steps = int(float(params.get("steps", 50000)))
    threads = int(float(params.get("threads", 1)))

    script = textwrap.dedent(
        f"""\
        #!/bin/bash

        cd {run_dir}
        rm -rf VTU_F TrajTSI
        mkdir -p TrajTSI
        ln -sf {mesh} TrajTSI/dts0.tsi

        {cmd} -in {input_file} \\
            -top {topol} \\
            -e {steps} \\
            -nt {threads} \\
            -seed 76532
        """
    )
    with open(run_dir / "run.sh", "w", encoding="utf-8") as f:
        f.write(script)


def _finalize_runs(
    screen_dir: Path,
    summary: Dict,
    topol_path: Optional[Path] = None,
) -> None:
    """Ensure every run has topol.top and run.sh, then write launcher scripts.

    Parameters
    ----------
    screen_dir : Path
        Screen output directory.
    summary : dict
        Screen summary (as returned by :func:`_write_summary`).
    topol_path : Path, optional
        Source topology file to copy into runs that lack one.
        When *None*, falls back to ``<screen_dir>/topol.top``.
    """
    if topol_path is None:
        topol_path = screen_dir / "topol.top"

    for run_info in summary["runs"]:
        run_dir = screen_dir / run_info["run_id"]
        topol_copy = run_dir / "topol.top"
        if not topol_copy.exists() and topol_path.exists():
            shutil.copy2(topol_path, topol_copy)
        _write_run_script(run_dir, run_info["parameters"])

    write_launcher_scripts(screen_dir, summary)


def generate_screen(
    mesh_conf: Optional[Dict],
    output_dir: str,
    mesh: str,
    volume_path: Optional[str] = None,
    hmff_params: Optional[Dict] = None,
    sim_params: Optional[Dict] = None,
    extra_config: str = "",
    coupling_params: Optional[Dict] = None,
    use_filters: bool = False,
    lowpass_cutoff: Optional[float] = None,
    highpass_cutoff: Optional[float] = None,
    plane_norm: Optional[str] = None,
) -> Dict:
    """Generate a DTS parameter screen.

    Parameters
    ----------
    mesh_conf : dict or None
        Mesh configuration with 'file', 'scale_factor', 'offset' keys.
        When *None*, ``mesh`` is treated as a direct file path and
        scale/offset default to ``1`` / ``0,0,0``.
    output_dir : str
        Base directory for screen output.
    mesh : str
        Selected mesh file from mesh_conf, or a direct path to a
        mesh file when *mesh_conf* is *None*.
    volume_path : str, optional
        Path to density volume.  When *None* a standard Helfrich
        energy method is used instead of HMFF.
    hmff_params : dict, optional
        Physical parameters (kappa, kappa0 and optionally xi,
        invert_contrast).  Screened values should already be
        ``{{name:range}}`` strings.
    sim_params : dict, optional
        Simulation parameters (steps, threads, temperature).
    extra_config : str
        Raw DTS config lines appended verbatim.
    coupling_params : dict, optional
        Coupling configurations keyed by coupling name.
    use_filters : bool
        Whether to apply bandpass filtering.
    lowpass_cutoff : float
        Lowpass filter cutoff.
    highpass_cutoff : float
        Highpass filter cutoff.
    plane_norm : str
        Axis normalization.

    Returns
    -------
    dict
        Summary with run count, parameter names, directory paths.
    """
    from pyfreedts.screen import ParameterParser

    hmff_params = hmff_params or {}
    sim_params = sim_params or {}

    warnings.warn(
        "Setup FreeDTS Screen - Citation: "
        "[1] Pezeshkian, W. et al. (2024) Nat. Commun., "
        "doi.org/10.1038/s41467-024-44819-w."
    )

    screen_dir = Path(output_dir)
    makedirs(screen_dir, exist_ok=True)
    report_progress(message="Preparing inputs", current=1, total=4)

    if mesh_conf is not None:
        mesh_index = mesh_conf["file"].index(mesh)
        mesh_offset = mesh_conf["offset"][mesh_index]
        mesh_scale = mesh_conf["scale_factor"][mesh_index]
    else:
        mesh_offset = "0,0,0"
        mesh_scale = "1"

    if volume_path is not None:
        volume_path = _prepare_volume(
            volume_path=volume_path,
            output_dir=output_dir,
            use_filters=use_filters,
            lowpass_cutoff=lowpass_cutoff,
            highpass_cutoff=highpass_cutoff,
            plane_norm=plane_norm,
        )

    report_progress(message="Building template", current=2, total=4)

    raw_template = _build_dts_template(
        volume_path=volume_path,
        mesh_scale=mesh_scale,
        mesh_offset=mesh_offset,
        hmff_params=hmff_params,
        sim_params=sim_params,
        extra_config=extra_config,
        coupling_params=coupling_params,
    )
    template_content, parameters = ParameterParser.parse_template(raw_template)

    template_path = screen_dir / "template.dts"
    template_path.write_text(template_content, encoding="utf-8")

    topol_path = screen_dir / "topol.top"
    topol_path.write_text(f"{mesh} 1\n", encoding="utf-8")

    report_progress(message="Generating runs", current=3, total=4)

    existing_runs = []
    merged_parameters = dict(parameters)
    summary_path = screen_dir / "screen_summary.json"
    if summary_path.exists():
        with open(summary_path, "r") as f:
            old_summary = json.load(f)
        existing_runs = old_summary["runs"]
        for name, info in old_summary.get("parameter_ranges", {}).items():
            old_vals = set(info.get("values", []))
            new_vals = set(merged_parameters.get(name, []))
            merged_parameters[name] = sorted(old_vals | new_vals)

    new_runs = _expand_screen(screen_dir, template_content, parameters, existing_runs)
    all_runs = existing_runs + new_runs
    summary = _write_summary(
        screen_dir, str(template_path), merged_parameters, all_runs
    )
    _finalize_runs(screen_dir, summary, topol_path)

    report_progress(message="Done", current=4, total=4)

    return {
        "total_runs": summary["total_runs"],
        "new_runs": len(new_runs),
        "parameter_names": summary["parameters"],
        "output_dir": output_dir,
        "summary_path": str(screen_dir / "screen_summary.json"),
    }


def write_launcher_scripts(output_dir: Path, summary: Dict) -> None:
    """Write local and SLURM launcher scripts for a screen.

    Generates ``run_all.sh`` for local execution (sequential, or via
    GNU parallel when available) and ``submit_slurm.sh`` for HPC
    submission as a SLURM array job.  Run scripts are discovered at
    runtime via ``run_*/run.sh`` globs so the launchers stay compact.

    Parameters
    ----------
    output_dir : Path
        Screen output directory containing per-run subdirectories.
    summary : dict
        Screen summary with ``runs`` and ``total_runs`` keys.
    """
    local_script = textwrap.dedent(
        """\
        #!/bin/bash
        set -euo pipefail

        SCREEN_DIR="$(cd "$(dirname "$0")" && pwd)"
        mapfile -t RUNS < <(find "$SCREEN_DIR" -path '*/run_*/run.sh' | sort)

        if command -v parallel &>/dev/null; then
            printf '%s\\n' "${RUNS[@]}" | parallel --bar bash {}
        else
            for script in "${RUNS[@]}"; do
                echo "Running $script"
                bash "$script"
            done
        fi
        """
    )
    local_path = output_dir / "run_all.sh"
    local_path.write_text(local_script, encoding="utf-8")
    local_path.chmod(0o755)

    n_runs = len(summary["runs"])
    if n_runs == 0:
        return
    slurm_script = textwrap.dedent(
        f"""\
        #!/bin/bash
        #SBATCH --job-name=dts_screen
        #SBATCH --array=0-{n_runs - 1}
        #SBATCH --ntasks=1
        #SBATCH --cpus-per-task=1
        #SBATCH --output=slurm_%A_%a.log

        SCREEN_DIR="$(cd "$(dirname "$0")" && pwd)"
        mapfile -t RUNS < <(find "$SCREEN_DIR" -path '*/run_*/run.sh' | sort)

        bash "${{RUNS[$SLURM_ARRAY_TASK_ID]}}"
        """
    )
    slurm_path = output_dir / "submit_slurm.sh"
    slurm_path.write_text(slurm_script, encoding="utf-8")
    slurm_path.chmod(0o755)


def extend_screen(screen_dir: str, new_screen_params: Dict[str, str]) -> Dict:
    """Extend an existing screen with new parameter combinations.

    Parameters
    ----------
    screen_dir : str
        Existing screen directory.
    new_screen_params : dict
        New parameter ranges to add.

    Returns
    -------
    dict
        Updated summary.
    """
    from pyfreedts.screen import ParameterParser

    screen_path = Path(screen_dir)

    with open(screen_path / "screen_summary.json", "r") as f:
        summary = json.load(f)

    template_content = (screen_path / "template.dts").read_text()

    new_values = {}
    for name, range_str in new_screen_params.items():
        _, parsed = ParameterParser.parse_template(f"{{{{placeholder:{range_str}}}}}")
        new_values[name] = parsed.get("placeholder", [])

    new_runs = _expand_screen(
        screen_path, template_content, new_values, summary["runs"]
    )

    all_runs = summary["runs"] + new_runs
    all_parameters = {}
    for name in set(
        list(summary.get("parameter_ranges", {}).keys()) + list(new_values.keys())
    ):
        existing_vals = (
            summary.get("parameter_ranges", {}).get(name, {}).get("values", [])
        )
        merged = sorted(set(existing_vals + new_values.get(name, [])))
        all_parameters[name] = merged

    updated_summary = _write_summary(
        screen_path,
        summary.get("template_file", ""),
        all_parameters,
        all_runs,
    )
    _finalize_runs(screen_path, updated_summary)

    return {
        "new_runs": len(new_runs),
        "total_runs": len(all_runs),
        "parameter_names": sorted(all_parameters.keys()),
    }


def get_screen_status(screen_dir: str) -> List[Dict]:
    """Check status of all runs in a screen directory.

    Parameters
    ----------
    screen_dir : str
        Screen directory path.

    Returns
    -------
    list of dict
        Each dict has run_id, parameters, status.
    """
    summary_path = Path(screen_dir) / "screen_summary.json"
    if not summary_path.exists():
        return []

    with open(summary_path, "r") as f:
        summary = json.load(f)

    results = []
    for run_info in summary["runs"]:
        run_dir = Path(screen_dir) / run_info["run_id"]
        results.append(
            {
                "run_id": run_info["run_id"],
                "parameters": run_info["parameters"],
                "status": run_status(run_dir),
            }
        )

    return results



"""
Batch rename dialog for renaming multiple geometry items at once.

Copyright (c) 2024 European Molecular Biology Laboratory

Author: Valentin Maurer <valentin.maurer@embl-hamburg.de>
"""

import re
from typing import List

from qtpy.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QGridLayout,
    QCheckBox,
    QStackedWidget,
    QWidget,
    QPushButton,
)

from ..stylesheets import (
    QPushButton_style,
    QLineEdit_style,
    QCheckBox_style,
    Colors,
)
from ..widgets import DialogFooter


class BatchRenameDialog(QDialog):
    """Dialog for batch renaming with Replace Text and Format modes.

    Modes
    -----
    Replace Text
        Find and replace within names. Supports plain text (default) and
        optional regular expressions for power users.
    Format
        Pattern template with ``{name}`` (original name), ``{i}`` (sequential
        index), and ``{i:03}`` (zero-padded index) tokens.

    Parameters
    ----------
    names : List[str]
        Current names of items in tree order.
    parent : QWidget, optional
        Parent widget.
    """

    MODES = ["Replace", "Format"]

    def __init__(self, names: List[str], parent=None):
        super().__init__(parent)
        self.names = names
        self.result_names = list(names)

        self.setWindowTitle("Rename")
        self.setMinimumWidth(480)

        self._setup_ui()
        self._connect_signals()
        self._update_result()

    def _setup_ui(self):
        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(20, 20, 20, 10)
        main_layout.setSpacing(0)

        # Mode tabs as segmented control
        tab_row = QHBoxLayout()
        tab_row.setSpacing(0)
        self.mode_buttons = []
        tab_container = QWidget()
        tab_layout = QHBoxLayout(tab_container)
        tab_layout.setContentsMargins(0, 0, 0, 0)
        tab_layout.setSpacing(0)
        for i, mode in enumerate(self.MODES):
            btn = QPushButton(mode)
            btn.setCheckable(True)
            btn.setChecked(i == 0)
            radius_l = "4px" if i == 0 else "0px"
            radius_r = "4px" if i == len(self.MODES) - 1 else "0px"
            margin_l = "0px" if i == 0 else "-1px"
            btn.setStyleSheet(
                f"""
                QPushButton {{
                    border: 1px solid {Colors.BORDER_DARK};
                    border-radius: 0px;
                    border-top-left-radius: {radius_l};
                    border-bottom-left-radius: {radius_l};
                    border-top-right-radius: {radius_r};
                    border-bottom-right-radius: {radius_r};
                    padding: 5px 18px;
                    font-size: 12px;
                    color: {Colors.TEXT_SECONDARY};
                    background: transparent;
                    margin-left: {margin_l};
                }}
                QPushButton:checked {{
                    background: {Colors.BG_TERTIARY};
                    color: {Colors.TEXT_PRIMARY};
                    font-weight: 500;
                }}
                QPushButton:hover:!checked {{
                    background: {Colors.BG_HOVER};
                }}
            """
            )
            btn.clicked.connect(lambda checked, idx=i: self._on_mode_clicked(idx))
            tab_layout.addWidget(btn)
            self.mode_buttons.append(btn)

        tab_row.addWidget(tab_container)
        tab_row.addStretch()
        main_layout.addLayout(tab_row)
        main_layout.addSpacing(14)

        # Stacked options (no group box)
        self.stack = QStackedWidget()
        self._build_replace_page()
        self._build_format_page()
        main_layout.addWidget(self.stack)
        main_layout.addSpacing(10)

        self.preview_label = QLabel()
        self.preview_label.setStyleSheet(
            f"color: {Colors.TEXT_MUTED}; font-size: 11px;"
        )
        main_layout.addWidget(self.preview_label)
        main_layout.addSpacing(8)

        footer = DialogFooter(dialog=self, margin=(0, 4, 0, 0))
        footer.accept_button.setText("Rename")
        main_layout.addWidget(footer)

        self.setStyleSheet(QPushButton_style)

    def _build_replace_page(self):
        page = QWidget()
        grid = QGridLayout(page)
        grid.setSpacing(10)
        grid.setContentsMargins(0, 0, 0, 0)

        find_label = QLabel("Find")
        find_label.setStyleSheet(f"color: {Colors.TEXT_SECONDARY}; font-size: 12px;")
        grid.addWidget(find_label, 0, 0)
        self.find_input = QLineEdit()
        self.find_input.setPlaceholderText("Text to find")
        self.find_input.setStyleSheet(QLineEdit_style)
        grid.addWidget(self.find_input, 0, 1)

        replace_label = QLabel("Replace")
        replace_label.setStyleSheet(f"color: {Colors.TEXT_SECONDARY}; font-size: 12px;")
        grid.addWidget(replace_label, 1, 0)
        self.replace_input = QLineEdit()
        self.replace_input.setPlaceholderText("Replacement (leave empty to delete)")
        self.replace_input.setStyleSheet(QLineEdit_style)
        grid.addWidget(self.replace_input, 1, 1)

        self.error_label = QLabel()
        self.error_label.setStyleSheet(
            f"color: {Colors.ERROR}; font-size: 10px; padding: 0;"
        )
        self.error_label.hide()
        grid.addWidget(self.error_label, 2, 1)

        options_row = QHBoxLayout()
        self.case_sensitive_check = QCheckBox("Case sensitive")
        self.case_sensitive_check.setChecked(True)
        self.case_sensitive_check.setStyleSheet(QCheckBox_style)
        options_row.addWidget(self.case_sensitive_check)

        self.regex_check = QCheckBox("Regular expression")
        self.regex_check.setStyleSheet(QCheckBox_style)
        options_row.addWidget(self.regex_check)
        options_row.addStretch()

        grid.addLayout(options_row, 3, 1)

        self.stack.addWidget(page)

    def _build_format_page(self):
        page = QWidget()
        grid = QGridLayout(page)
        grid.setSpacing(10)
        grid.setContentsMargins(0, 0, 0, 0)

        template_label = QLabel("Template")
        template_label.setStyleSheet(
            f"color: {Colors.TEXT_SECONDARY}; font-size: 12px;"
        )
        grid.addWidget(template_label, 0, 0)
        self.pattern_input = QLineEdit()
        self.pattern_input.setPlaceholderText("e.g. Mitochondrion {i}")
        self.pattern_input.setText("{name}")
        self.pattern_input.setStyleSheet(QLineEdit_style)
        self.pattern_input.selectAll()
        grid.addWidget(self.pattern_input, 0, 1)

        token_help = QLabel("{name} = original · {i} = number · {i:03} = zero-padded")
        token_help.setStyleSheet(f"color: {Colors.TEXT_MUTED}; font-size: 10px;")
        grid.addWidget(token_help, 1, 1)

        self.stack.addWidget(page)

    def _connect_signals(self):
        self.find_input.textChanged.connect(self._update_result)
        self.replace_input.textChanged.connect(self._update_result)
        self.case_sensitive_check.stateChanged.connect(self._update_result)
        self.regex_check.stateChanged.connect(self._update_result)
        self.pattern_input.textChanged.connect(self._update_result)

    def _on_mode_clicked(self, index):
        for i, btn in enumerate(self.mode_buttons):
            btn.setChecked(i == index)
        self.stack.setCurrentIndex(index)
        self._update_result()

    def _compute_names(self):
        if self.stack.currentIndex() == 0:
            return self._apply_replace()
        return self._apply_format()

    def _apply_replace(self):
        find = self.find_input.text()
        replace = self.replace_input.text()
        if not find:
            return list(self.names)

        use_regex = self.regex_check.isChecked()
        case_sensitive = self.case_sensitive_check.isChecked()
        flags = 0 if case_sensitive else re.IGNORECASE

        result = []
        for name in self.names:
            try:
                if use_regex:
                    new_name = re.sub(find, replace, name, flags=flags)
                else:
                    new_name = re.sub(re.escape(find), replace, name, flags=flags)
                self.error_label.hide()
            except re.error as e:
                self.error_label.setText(f"Invalid pattern: {e}")
                self.error_label.show()
                return list(self.names)
            result.append(new_name)
        return result

    def _apply_format(self):
        pattern = self.pattern_input.text()

        result = []
        for idx, name in enumerate(self.names):
            text = pattern
            text = re.sub(
                r"\{i:(\d+)\}",
                lambda m: str(idx + 1).zfill(int(m.group(1))),
                text,
            )
            text = text.replace("{i}", str(idx + 1))
            text = text.replace("{name}", name)
            result.append(text)
        return result

    def _update_result(self):
        self.result_names = self._compute_names()
        if self.names[0] != self.result_names[0]:
            self.preview_label.setText(
                f"Example: {self.names[0]}  →  {self.result_names[0]}"
            )
        else:
            self.preview_label.setText("")



from qtpy.QtWidgets import (
    QVBoxLayout,
    QHBoxLayout,
    QDialog,
    QLabel,
    QComboBox,
    QPushButton,
    QWidget,
    QDoubleSpinBox,
    QScrollArea,
    QGroupBox,
    QFrame,
    QLineEdit,
    QMessageBox,
    QCheckBox,
)
import qtawesome as qta

from ..widgets import DialogFooter
from ..stylesheets import QPushButton_style, Colors


class MeshMappingRow(QWidget):
    def __init__(self, clusters, is_first=False, parent=None, dialog=None):
        super().__init__(parent)
        self.clusters = clusters
        self.dialog = dialog
        self.setup_ui()
        self.update_button_state(is_first)

    def setup_ui(self):
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)

        self.name_edit = QLineEdit()
        self.name_edit.setPlaceholderText("Protein Name")

        self.cluster_combo, names = QComboBox(), []
        for name, data in self.clusters:
            names.append(name)
            self.cluster_combo.addItem(name, data)

        self.name_edit.setText(self.cluster_combo.currentText())
        self.cluster_combo.currentTextChanged.connect(self.name_edit.setText)

        self.toggle_btn = QPushButton()
        self.toggle_btn.setFixedWidth(20)

        layout.addWidget(self.name_edit, 1)
        layout.addWidget(self.cluster_combo, 2)
        layout.addWidget(self.toggle_btn)

    def update_button_state(self, state):
        if state:
            self.toggle_btn.setIcon(qta.icon("ph.plus", color=Colors.ICON))
            self.toggle_btn.clicked.connect(self.add_requested)
            return None

        self.toggle_btn.setIcon(qta.icon("ph.trash", color=Colors.ICON))
        self.toggle_btn.clicked.connect(self.deleteLater)

    def add_requested(self):
        if self.dialog and hasattr(self.dialog, "add_mapping_row"):
            self.dialog.add_mapping_row()

    def get_parameters(self):
        return {
            "name": self.name_edit.text().strip(),
            "data": self.cluster_combo.currentData(),
        }

    def is_valid(self):
        return bool(self.name_edit.text().strip() and self.cluster_combo.currentText())


class MeshMappingDialog(QDialog):
    def __init__(self, fits, clusters, parent=None):
        super().__init__(parent)
        self.fits = fits
        self.clusters = clusters

        self.setWindowTitle("Backmapping")
        self.resize(500, 540)
        self.setup_ui()
        self.setStyleSheet(QPushButton_style)

    def setup_ui(self):
        from ..icons import dialog_margin, footer_margin

        layout = QVBoxLayout(self)
        layout.setSpacing(15)
        layout.setContentsMargins(*dialog_margin)

        config_group = QGroupBox("Mesh")
        config_layout = QVBoxLayout()
        config_layout.setSpacing(10)

        fit_layout = QHBoxLayout()
        fit_label = QLabel("Surface:")
        self.fit_combo = QComboBox()
        for name, data in self.fits:
            self.fit_combo.addItem(name, data)
        fit_layout.addWidget(fit_label)
        fit_layout.addWidget(self.fit_combo)
        config_layout.addLayout(fit_layout)

        edge_layout = QHBoxLayout()
        edge_label = QLabel("Output Edge Length:")
        self.edge_length = QDoubleSpinBox()
        self.edge_length.setValue(40.0)
        self.edge_length.setRange(0, 1e32)
        self.edge_length.setSingleStep(0.1)
        edge_layout.addWidget(edge_label)
        edge_layout.addWidget(self.edge_length)
        config_layout.addLayout(edge_layout)

        config_group.setLayout(config_layout)
        layout.addWidget(config_group)

        projection_group = QGroupBox("Projection")
        projection_layout = QVBoxLayout(projection_group)
        projection_layout.setSpacing(10)
        self.cast_rays = QCheckBox("Use raycasting to project proteins onto mesh.")
        self.cast_rays.toggled.connect(
            lambda: self.flip_normals.setEnabled(self.cast_rays.isChecked())
        )

        self.flip_normals = QCheckBox("Flip normal direction (needs to point inward).")
        self.cast_rays.setChecked(False)
        self.flip_normals.setChecked(False)
        self.flip_normals.setEnabled(False)
        projection_layout.addWidget(self.cast_rays)
        projection_layout.addWidget(self.flip_normals)
        layout.addWidget(projection_group)

        mapping_group = QGroupBox("Inclusions")
        mapping_layout = QVBoxLayout()

        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setFrameShape(QFrame.Shape.NoFrame)

        self.mapping_container = QWidget()
        self.mapping_layout = QVBoxLayout(self.mapping_container)
        self.mapping_layout.setContentsMargins(0, 0, 0, 0)
        self.mapping_layout.setSpacing(5)
        self.mapping_layout.addStretch()

        scroll.setWidget(self.mapping_container)
        mapping_layout.addWidget(scroll)
        mapping_group.setLayout(mapping_layout)
        layout.addWidget(mapping_group)

        footer = DialogFooter(dialog=self, margin=footer_margin)
        layout.addWidget(footer)

        first_row = MeshMappingRow(
            self.clusters, is_first=True, parent=self.mapping_container, dialog=self
        )
        self.mapping_layout.insertWidget(self.mapping_layout.count() - 1, first_row)

    def accept(self):
        """Validate the dialog inputs before accepting"""
        if not self.fit_combo.currentText():
            QMessageBox.warning(
                self, "Validation Error", "Please select a surface fit."
            )
            return

        valid_mappings = False
        for i in range(self.mapping_layout.count() - 1):
            widget = self.mapping_layout.itemAt(i).widget()
            if isinstance(widget, MeshMappingRow) and widget.is_valid():
                valid_mappings = True
                break

        if not valid_mappings:
            QMessageBox.warning(
                self,
                "Validation Error",
                "Please add at least one valid cluster mapping.",
            )
            return

        return super().accept()

    def add_mapping_row(self):
        new_row = MeshMappingRow(
            self.clusters, is_first=False, parent=self.mapping_container, dialog=self
        )
        self.mapping_layout.insertWidget(self.mapping_layout.count() - 1, new_row)

    def get_parameters(self):
        selected_fit = self.fit_combo.currentData()
        edge_length = self.edge_length.value()

        mappings = []
        for i in range(self.mapping_layout.count() - 1):
            widget = self.mapping_layout.itemAt(i).widget()
            if isinstance(widget, MeshMappingRow):
                mappings.append(widget.get_parameters())

        return (
            selected_fit,
            edge_length,
            mappings,
            self.cast_rays.isChecked(),
            self.flip_normals.isChecked(),
        )
